<?php

namespace Drupal\genoring_jbrowse\EventSubscriber;

use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\genoring\Event\DataModelManagementEvent;
use Drupal\genoring\Event\GenoringEvents;
use Drupal\menu_link_content\Entity\MenuLinkContent;
use GuzzleHttp\Exception\GuzzleException;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * GenoRing event subscriber to handle default behaviors.
 */
class JbrowseEventSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * Constructs a GenoringEventsSubscriber object.
   *
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   */
  public function __construct(TranslationInterface $string_translation) {
    $this->stringTranslation = $string_translation;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[GenoringEvents::DATA_MODEL_MANAGEMENT][] = 'jbrowseDataModelManagement';
    return $events;
  }

  /**
   * Provides default action to available data models.
   *
   * @param \Drupal\genoring\Event\DataModelManagementEvent $event
   *   The event.
   */
  public function jbrowseDataModelManagement(DataModelManagementEvent $event) {
    // @todo Implement.
    $data_model = $event->getDataModel();
    $management_form = $event->getManagementForm();
    // JBrowse supports the 'assembly' data model.
    if ('__assembly' == strstr($data_model->id(), '__')) {
      $assembly_id = strstr($data_model->id(), '__', TRUE);
      // Check if GenoRing Tools is available.
      $genoring_prefix = getenv('COMPOSE_PROJECT_NAME') ?: 'genoring';
      $fs = @fsockopen($genoring_prefix . '-genoringtools', '80', $errno, $errstr, 2);
      if (!empty($fs)) {
        // GenoRing Tools available, we can process.
        fclose($fs);
        // Get all assemblies.
        $assembly_storage = \Drupal::entityTypeManager()->getStorage($data_model->id());
        $assembly_ids = $assembly_storage->getQuery()->accessCheck(FALSE)->execute();
        $assemblies = $assembly_storage->loadMultiple($assembly_ids);
        $assembly_files = [];
        foreach ($assemblies as $assembly) {
          $assembly_files[] = [
            'fastas' => array_map(
              function ($item) {
                return $item['value'] ?? '';
              },
              $assembly->get('genoring_genome_fasta')->getValue()
            ),
            'gff3s' => array_map(
              function ($item) {
                return $item['value'] ?? '';
              },
              $assembly->get('genoring_genome_gff3')->getValue()
            ),
          ];
        }
        // Currently, we only have one assembly. But in the future, we may want
        // to load multiple assemblies or tracks.
        $fasta_uri = $assembly_files[0]['fastas'][0];
        $gff3_uri = $assembly_files[0]['gff3s'][0];
        // Make sure we got a FASTA and a GFF3.
        if (!empty($fasta_uri) && !empty($gff3_uri)) {
          $management_form['jbrowse'] = [
            '#type' => 'fieldset',
            '#title' => $this->t('JBrowse'),
            'assembly' => [
              '#type' => 'hidden',
              '#value' => $assembly_id,
            ],
            'fasta' => [
              '#type' => 'hidden',
              '#value' => $fasta_uri,
            ],
            'gff3' => [
              '#type' => 'hidden',
              '#value' => $gff3_uri,
            ],
            'load' => [
              '#type' => 'submit',
              '#value' => $this->t('Load assembly in JBrowse'),
              '#submit' => [[$this, 'submitJbrowseLoading']],
            ],
          ];
          $event->setManagementForm($management_form);
        }
      }
      else {
        // GenoRing Tools not available.
        $management_form['jbrowse'] = [
          '#type' => 'fieldset',
          '#title' => $this->t('JBrowse'),
          'info' => [
            '#type' => 'item',
            '#markup' => t('GenoRing Tools service is not available. Please restart GenoRing in "backend" mode for backend operations.'),
          ],
        ];
        $event->setManagementForm($management_form);
      }
    }
  }

  /**
   * Load data into JBrowse.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitJbrowseLoading(array &$form, FormStateInterface $form_state) {
    // Get current context.
    $dataset = $form_state->getFormObject()->getEntity();
    $assembly = strtolower(
      preg_replace(
        '/\W+/',
        '_',
        $form_state->getValue(
          ['data_models', 'assembly', 'jbrowse', 'assembly']
        )
      )
    );
    if (!$assembly) {
      // @todo Log and display an error.
      return;
    }
    $file_system = \Drupal::service('file_system');
    $data_manager = \Drupal::service('genoring.data_manager');
    $data_config = $data_manager->getConfiguration();
    $fasta_file_uri = $form_state->getValue(['data_models', 'assembly', 'jbrowse', 'fasta']);
    $gff3_file_uri = $form_state->getValue(['data_models', 'assembly', 'jbrowse', 'gff3']);

    // Locate JBrowse files in a dedicated directory.
    $jbrowse_assembly_uri = 'private://jbrowse/' . $assembly . '/';

    // FASTA file.
    $fasta_file_path = $file_system->realpath($fasta_file_uri);
    $fasta_filename = $file_system->basename($fasta_file_uri);
    // GFF3 file.
    $gff3_file_path = $file_system->realpath($gff3_file_uri);
    $gff3_filename = $file_system->basename($gff3_file_uri);

    // Only work if files exist.
    if (!empty($fasta_file_path) && !empty($gff3_file_path)) {
      // Generate FAI file.
      $fasta_metadata = $data_manager->getMetadata($fasta_file_uri);
      $fai_metadata = $fasta_metadata;
      $fai_metadata['format'] = 'fai';
      $fai_metadata['type'] = 'file';
      $fai_metadata['filename'] = $fasta_filename . '.fai';
      $fai_metadata['isSource'] = FALSE;
      unset($fai_metadata['dataset']);
      $fai_metadata['source'] = [$fasta_file_uri];
      $fai_file_uri = $data_manager->locateUri($fai_metadata);
      $fai_file_path = $data_manager->prepareFileDirectory($fai_file_uri, TRUE);
      $this->runFaidx($fasta_file_path, $fai_file_path);
      $data_manager->setMetadata($fai_file_uri, $fai_metadata);

      // Generate sorted GFF3 file.
      $gff3_metadata = $data_manager->getMetadata($gff3_file_uri);
      $sortedgff3_metadata = $gff3_metadata;
      $sortedgff3_metadata['format'] = 'gff3';
      $sortedgff3_metadata['type'] = 'file';
      $sortedgff3_metadata['filename'] = preg_replace('/\.gff3?$/i', '', $gff3_filename) . '.sorted.gff3';
      $sortedgff3_metadata['isSource'] = FALSE;
      unset($sortedgff3_metadata['dataset']);
      $sortedgff3_metadata['source'] = [$gff3_file_uri];
      $sortedgff3_metadata['tags'] = ['sorted'];
      $sortedgff3_file_uri = $data_manager->locateUri($sortedgff3_metadata);
      $sortedgff3_file_path = $data_manager->prepareFileDirectory($sortedgff3_file_uri, TRUE);
      $this->runGff3Sort($gff3_file_path, $sortedgff3_file_uri, $file_system);
      $this->runBgzip($sortedgff3_file_path);
      $this->runTabix($sortedgff3_file_path . '.gz');

      // Symlink files.
      $fasta_file = $data_manager->generateSymlink($fasta_file_uri, $jbrowse_assembly_uri . '/seq/' . $fasta_filename);
      $fai_file = $data_manager->generateSymlink($fai_file_uri, $jbrowse_assembly_uri . '/seq/' . $fai_metadata['filename']);
      $sorted_gff3_file = $data_manager->generateSymlink($sortedgff3_file_uri . '.gz', $jbrowse_assembly_uri . '/gff3/' . $sortedgff3_metadata['filename'] . '.gz');
      $gff3_index_file = $data_manager->generateSymlink($sortedgff3_file_uri . '.gz.tbi', $jbrowse_assembly_uri . '/gff3/' . $sortedgff3_metadata['filename'] . '.gz.tbi');

      // Generate tracks.conf.
      $tracks_conf = "[general]\nrefSeqs=seq/" . $fai_metadata['filename'] . "\n";
      $tracks_conf_uri = $jbrowse_assembly_uri . '/tracks.conf';
      $file_system->saveData($tracks_conf, $tracks_conf_uri, FileExists::Replace);
      // Generate trackList.json.
      $track_list_json = "{\n  \"tracks\": [\n    {\n      \"urlTemplate\": \"seq/$fasta_filename\",\n      \"storeClass\": \"JBrowse/Store/SeqFeature/IndexedFasta\",\n      \"type\": \"Sequence\",\n      \"label\": \"Chromosome_track\",\n      \"key\": \"chromosomes\"\n    },\n"
        . "    {\n      \"urlTemplate\": \"gff3/"
        . $sortedgff3_metadata['filename']
        . ".gz\",\n      \"storeClass\": \"JBrowse/Store/SeqFeature/GFF3Tabix\",\n      \"type\": \"CanvasFeatures\",\n      \"label\": \"Gene_track\",\n      \"key\": \"genes\"\n    }\n  ]\n}\n";
      $track_list_json_uri = $jbrowse_assembly_uri . '/trackList.json';
      $file_system->saveData($track_list_json, $track_list_json_uri, FileExists::Replace);

      // Add corresponding link to menu.
      $host = '://'
        . (getenv('GENORING_HOST') ?: 'localhost')
        . ((getenv('GENORING_PORT') ? ':' . getenv('GENORING_PORT') : ''))
        . '/';
      $assembly_uri = str_replace(
        '://default/',
        $host,
        \Drupal::urlGenerator()->generateFromRoute('<front>', [], ['absolute' => TRUE])
        . "jbrowse/index.html?data=data/$assembly"
      );
      $menuitem = \Drupal::entityTypeManager()
        ->getStorage('menu_link_content')
        ->loadByProperties(['link.uri' => $assembly_uri]);
      if (empty($menuitem)) {
        $menuitem = [
          'title' => "JBrowse $assembly",
          'link' => [
            'uri' => $assembly_uri,
          ],
          'menu_name' => 'main',
          'expanded' => TRUE,
        ];
        MenuLinkContent::create($menuitem)->save();
      }

    }
    else {
      // Debug.
      \Drupal::messenger()->addWarning('JBrowse loading: Missing data files.');
    }
    $form_state->disableRedirect();
  }

  /**
   * Generate an indexed FASTA.
   *
   * @param string $input_fasta_path
   *   Path (not a URI) to input FASTA file.
   * @param string $output_fai_path
   *   Path (not a URI) to output FAI file.
   */
  public function runFaidx(string $input_fasta_path, string $output_fai_path) {
    // Prepare GenoRing Tools server URL.
    $genoring_prefix = getenv('COMPOSE_PROJECT_NAME') ?: 'genoring';
    $genoring_tools_cgi = 'http://' . $genoring_prefix . '-genoringtools/cgi-bin/';
    // @todo Check execution.
    $http_client = \Drupal::service('http_client');
    try {
      $req_parameters = [
        'headers' => [
          'Accept' => 'application/json',
        ],
        'query' => [
          'fasta' => $input_fasta_path,
          'output' => $output_fai_path,
        ],
      ];
      $response = $http_client->request(
        'GET',
        $genoring_tools_cgi . 'faidx',
        $req_parameters
      );
      if ($response->getStatusCode() == 200) {

        $body = $response->getBody() . '';
        $data = json_decode($body, TRUE);
        // @todo Set metadata.
        // @todo Log execution result.
        // Debug.
        \Drupal::logger('genoring_jbrowse')->debug('faidx output: {placeholder}', ['placeholder' => print_r($data, TRUE)]);
      }
      else {
        // Not supported (2xx != 200 and others).
        // @todo Adjust error.
      }
    }
    catch (GuzzleException $exception) {
      // @todo Adjust error.
      \Drupal::logger('genoring_jbrowse')->error('Error loading tracks. Message: ' . $exception->getMessage());
    }
  }

  /**
   * Generate a sorted GFF3 file.
   *
   * @param string $input_gff3_path
   *   Path (not a URI) to GFF3 file.
   * @param string $output_sorted_gff3_uri
   *   URI (not a path) to output file.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   A file system service.
   */
  public function runGff3Sort(
    string $input_gff3_path,
    string $output_sorted_gff3_uri,
    FileSystemInterface $file_system,
  ) {
    // Prepare GenoRing Tools server URL.
    $genoring_prefix = getenv('COMPOSE_PROJECT_NAME') ?: 'genoring';
    $genoring_tools_cgi = 'http://' . $genoring_prefix . '-genoringtools/cgi-bin/';
    // @todo Check execution.
    $http_client = \Drupal::service('http_client');
    try {
      $req_parameters = [
        'headers' => [
          'Accept' => 'application/json',
        ],
        'query' => [
          'param' => $input_gff3_path,
        ],
      ];
      $response = $http_client->request(
        'GET',
        $genoring_tools_cgi . 'gff3sort',
        $req_parameters
      );
      if ($response->getStatusCode() == 200) {
        $body = $response->getBody() . '';
        $data = json_decode($body, TRUE);
        // Save result.
        $file_system->saveData($data['output'], $output_sorted_gff3_uri, FileExists::Replace);
        // @todo No metadata here as it will be compressed.
        // @todo Log execution result.
        $data['output'] = substr($data['output'] ?? '', 0, 50) . '...';
        \Drupal::logger('genoring_jbrowse')->debug('gff3sort output: {placeholder}', ['placeholder' => print_r($data, TRUE)]);
      }
      else {
        // Not supported (2xx != 200 and others).
        // @todo Adjust error.
      }
    }
    catch (GuzzleException $exception) {
      // @todo Adjust error.
      \Drupal::logger('genoring_jbrowse')->error('Error loading tracks. Message: ' . $exception->getMessage());
    }
  }

  /**
   * BGZip a file.
   *
   * @param string $input_file_path
   *   Input file path (not a URI!) of the file that should be compressed.
   */
  public function runBgzip(
    string $input_file_path,
  ) {
    // Prepare GenoRing Tools server URL.
    $genoring_prefix = getenv('COMPOSE_PROJECT_NAME') ?: 'genoring';
    $genoring_tools_cgi = 'http://' . $genoring_prefix . '-genoringtools/cgi-bin/';
    // @todo Check execution.
    $http_client = \Drupal::service('http_client');
    try {
      $req_parameters = [
        'headers' => [
          'Accept' => 'application/json',
        ],
        'query' => [
          'param' => $input_file_path,
        ],
      ];
      $response = $http_client->request(
        'GET',
        $genoring_tools_cgi . 'bgzip',
        $req_parameters
      );
      if ($response->getStatusCode() == 200) {
        $body = $response->getBody() . '';
        $data = json_decode($body, TRUE);
        // @todo Set metadata.
        // @todo Log execution result.
        \Drupal::logger('genoring_jbrowse')->debug('bgzip output: {placeholder}', ['placeholder' => print_r($data, TRUE)]);
      }
      else {
        // Not supported (2xx != 200 and others).
        // @todo Adjust error.
      }
    }
    catch (GuzzleException $exception) {
      // @todo Adjust error.
      \Drupal::logger('genoring_jbrowse')->error('Error loading tracks. Message: ' . $exception->getMessage());
    }
  }

  /**
   * Index a compressed GFF3 file.
   *
   * @param string $input_file_path
   *   Path (not a URI!) to input BGZip file file.
   */
  public function runTabix(
    string $input_file_path,
  ) {
    // Prepare GenoRing Tools server URL.
    $genoring_prefix = getenv('COMPOSE_PROJECT_NAME') ?: 'genoring';
    $genoring_tools_cgi = 'http://' . $genoring_prefix . '-genoringtools/cgi-bin/';
    // @todo Check execution.
    $http_client = \Drupal::service('http_client');
    try {
      $req_parameters = [
        'headers' => [
          'Accept' => 'application/json',
        ],
        'query' => [
          'param' => "-p gff $input_file_path",
        ],
      ];
      $response = $http_client->request(
        'GET',
        $genoring_tools_cgi . 'tabix',
        $req_parameters
      );
      if ($response->getStatusCode() == 200) {
        $body = $response->getBody() . '';
        $data = json_decode($body, TRUE);
        // @todo Set metadata.
        // @todo Log execution result.
        \Drupal::logger('genoring_jbrowse')->debug('tabix output: {placeholder}', ['placeholder' => print_r($data, TRUE)]);
      }
      else {
        // Not supported (2xx != 200 and others).
        // @todo Adjust error.
      }
    }
    catch (GuzzleException $exception) {
      // @todo Adjust error.
      \Drupal::logger('genoring_jbrowse')->error('Error loading tracks. Message: ' . $exception->getMessage());
    }
  }

}
