<?php

namespace Drupal\genoring\DataManager;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Render\Element\Email;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Url;
use Drupal\external_entities\Plugin\ExternalEntities\DataAggregator\GroupAggregator;
use Drupal\file\Entity\File;
use Drupal\file\FileInterface;
use Drupal\file\FileRepository;
use Drupal\genoring\Event\DataModelsEvent;
use Drupal\genoring\Event\DataTypeMetadataEvent;
use Drupal\genoring\Event\DataTypeOperationsEvent;
use Drupal\genoring\Event\FileTypesEvent;
use Drupal\genoring\Event\GenoringEvents;
use Drupal\genoring\Exception\DataManagerException;
use Drupal\genoring\Exception\MetadataException;
use Drupal\link\LinkItemInterface;
use GuzzleHttp\ClientInterface;
use JsonPath\InvalidJsonException;
use JsonPath\InvalidJsonPathException;
use JsonPath\JsonObject;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Yaml\Exception\ParseException;
use Symfony\Component\Yaml\Yaml;

/**
 * A base class for data locators.
 */
class DataManager implements DataManagerInterface, ConfigurableInterface, PluginFormInterface {

  use StringTranslationTrait;

  /**
   * Data manager configuration key.
   */
  const CONFIG_KEY = 'genoring.data_manager';

  /**
   * Cache key for data model definitions.
   */
  const CACHE_KEY_DATA_MODELS = 'genoring.data_models';

  /**
   * Cache key for supported file types.
   */
  const CACHE_KEY_FILE_TYPES = 'genoring.file_types';

  /**
   * Cache key for metadata fields.
   */
  const CACHE_KEY_TYPE_METADATA = 'genoring.type_metadata';

  /**
   * Cache key for available pre-processing operations.
   */
  const CACHE_KEY_TYPE_OPERATIONS = 'genoring.type_operations';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The file repository service.
   *
   * @var \Drupal\file\FileRepository
   */
  protected $fileRepository;

  /**
   * File system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Default data manager configuration.
   *
   * @var array
   */
  public static $defaultConfiguration = [
    'upload_uri' => 'private://upload',
    'data_uri' => 'private://genoring',
    'download_uri' => 'private://download',
    'data_locators' => [
      [
        'id' => 'default',
        'config' => [
          'data_uri' => 'private://genoring/',
          'replace' => 1,
        ],
        'rules' => '',
        'notes' => 'Data root directory.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => 'src/',
          'term' => 'format',
        ],
        'rules' => "-format=metadata\n+format,isSource=true",
        'notes' => 'Places source files in "src" directory, sorted by file type.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => 'src/other/',
          'term' => '',
        ],
        'rules' => "-format\n+isSource=true",
        'notes' => 'Places other source files without format in "src/other" directory.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => 'gen/',
          'term' => '',
        ],
        'rules' => "-format=metadata\n-isSource=true\n-dataset",
        'notes' => 'Places generated files in "gen" directory.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => 'dts/',
          'term' => '',
        ],
        'rules' => "-format=metadata\n-isSource=true\n+dataset",
        'notes' => 'Places dataset files or symlinks in "dts" directory.',
      ],
      [
        'id' => 'taxonomy',
        'config' => [
          'mode' => 'code',
        ],
        'rules' => "-format=metadata",
        'notes' => 'Sorts files by species codes.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => '',
          'term' => 'dataset',
        ],
        'rules' => "-format=metadata\n-isSource=true\n+dataset",
        'notes' => 'Places dataset files or symlinks in the directory named using the dataset name.',
      ],
      [
        'id' => 'term',
        'config' => [
          'data_uri' => '',
          'term' => 'assembly',
        ],
        'rules' => "-format=metadata\n+assembly",
        'notes' => 'Sorts files by assembly names.',
      ],
      [
        'id' => 'metadata',
        'config' => [
          'data_uri' => '.metadata',
        ],
        'rules' => "+format=metadata,targetUri",
        'notes' => 'Places metadata files.',
      ],
    ],
  ];

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerChannelFactory;

  /**
   * The data locator plugin logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannel
   */
  protected $logger;

  /**
   * The cache backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The data manager configuration.
   *
   * @var array
   */
  protected $configuration;

  /**
   * The data locators.
   *
   * @var \Drupal\genoring\DataLocator\DataLocatorInterface[]
   */
  protected $dataLocators;

  /**
   * Default GenoRing field definition array.
   *
   * @var array
   * @see \Drupal\genoring\DataManager\DataManagerInterface
   * @see https://www.drupal.org/docs/drupal-apis/entity-api/fieldtypes-fieldwidgets-and-fieldformatters
   */
  protected $dataFieldDefinitions = [
    'types' => [
      'boolean',
      // Orignal name: 'datetime'.
      'date',
      'daterange',
      'email',
      'float',
      'genoring_input',
      // @todo Add 'image'.
      'integer',
      'link',
      // Orignal name: 'entity_reference'.
      'reference',
      'string',
      // Orignal name: 'text_long'.
      'text',
      // Derived from 'text_long'.
      'text_formatted',
      'timestamp',
      'telephone',
    ],
    'form_element' => [
      'boolean' => [
        '#type' => 'checkbox',
      ],
      'date' => [
        '#type' => 'datetime',
        '#date_year_range' => '1902:2037',
      ],
      'daterange' => [
        // @todo Needs to be adjusted.
        '#type' => 'container',
        'from' => [
          '#type' => 'textfield',
          '#size' => 60,
          '#maxlength' => 255,
        ],
        'to' => [
          '#type' => 'textfield',
          '#size' => 60,
          '#maxlength' => 255,
        ],
      ],
      'email' => [
        '#type' => 'email',
        '#size' => 60,
        '#maxlength' => Email::EMAIL_MAX_LENGTH,
      ],
      'float' => [
        '#type' => 'number',
        '#step' => 'any',
      ],
      'genoring_input' => [
        '#type' => 'genoring_input',
      ],
      // @todo Needs to be adjusted.
      // @code
      // 'image' => [
      //   '#type' => 'textfield',
      //   '#size' => 60,
      //   '#maxlength' => 255,
      // ],
      // @endcode
      'integer' => [
        '#type' => 'number',
      ],
      'license' => [
        '#type' => 'genoring_license',
      ],
      'link' => [
        // @todo Needs to be adjusted.
        '#type' => 'url',
        '#size' => 100,
        '#maxlength' => 2048,
      ],
      'reference' => [
        /*
        @todo Needs to be adjusted.
        @code
        '#type' => 'entity_autocomplete',
        '#target_type' => ...,
        '#selection_handler' => ...,
        '#selection_settings' => [
          'match_operator' => 'CONTAINS',
          'match_limit' => 10,
        ],
        '#validate_reference' => FALSE,
        '#maxlength' => 1024,
        '#size' => 60,
        @endcode
        */
        '#type' => 'textfield',
        '#size' => 60,
        '#maxlength' => 255,
      ],
      'string' => [
        '#type' => 'textfield',
        '#size' => 60,
        '#maxlength' => 255,
      ],
      'taxonomy' => [
        '#type' => 'genoring_taxonomy',
      ],
      'text' => [
        '#type' => 'textarea',
        '#rows' => 5,
        '#attributes' => ['class' => ['js-text-full', 'text-full']],
      ],
      'text_formatted' => [
        '#type' => 'textarea',
        '#rows' => 5,
        '#attributes' => ['class' => ['js-text-full', 'text-full']],
      ],
      'timestamp' => [
        // @todo Needs to be adjusted.
        '#type' => 'textfield',
        '#size' => 60,
        '#maxlength' => 255,
      ],
      'telephone' => [
        // @todo Needs to be adjusted.
        '#type' => 'textfield',
        '#size' => 60,
        '#maxlength' => 255,
      ],
    ],
    'field_config' => [
      'boolean' => [
        'label' => 'Boolean value',
        'description' => '',
        'field_type' => 'boolean',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'on_label' => 'On',
          'off_label' => 'Off',
        ],
      ],
      'comment' => [
        'label' => 'Comment',
        'description' => '',
        'field_type' => 'comment',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [
          'default_mode' => 1,
          'per_page' => 50,
          'anonymous' => 0,
          'form_location' => TRUE,
          'preview' => 1,
        ],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'date' => [
        'label' => 'Date and time',
        'description' => '',
        'field_type' => 'datetime',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'daterange' => [
        'label' => 'Date Range',
        'description' => '',
        'field_type' => 'daterange',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [],
      ],
      'decimal' => [
        'label' => 'Decimal',
        'description' => '',
        'field_type' => 'decimal',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'min' => NULL,
          'max' => NULL,
          'prefix' => '',
          'suffix' => '',
        ],
      ],
      'email' => [
        'label' => 'Email',
        'description' => '',
        'field_type' => 'email',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'reference' => [
        'label' => 'Entity Reference',
        'description' => '',
        'field_type' => 'entity_reference',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [
          'handler' => 'default:node',
          'handler_settings' => [
            'target_bundles' => [
              'page' => 'page',
            ],
            'sort' => [
              'field' => '_none',
              'direction' => 'ASC',
            ],
            'auto_create' => FALSE,
            'auto_create_bundle' => '',
          ],
        ],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'file' => [
        'label' => 'File',
        'description' => '',
        'field_type' => 'file',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'handler' => 'default:file',
          'handler_settings' => [],
          'file_directory' => '[date:custom:Y]-[date:custom:m]',
          'file_extensions' => 'txt',
          'max_filesize' => '',
          'description_field' => FALSE,
        ],
      ],
      'float' => [
        'label' => 'Float',
        'description' => '',
        'field_type' => 'float',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'min' => NULL,
          'max' => NULL,
          'prefix' => '',
          'suffix' => '',
        ],
      ],
      'image' => [
        'label' => 'Image',
        'description' => '',
        'field_type' => 'image',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'handler' => 'default:file',
          'handler_settings' => [],
          'file_directory' => '[date:custom:Y]-[date:custom:m]',
          'file_extensions' => 'png gif jpg jpeg webp',
          'max_filesize' => '',
          'max_resolution' => '',
          'min_resolution' => '',
          'alt_field' => TRUE,
          'alt_field_required' => TRUE,
          'title_field' => FALSE,
          'title_field_required' => FALSE,
          'default_image' => [
            'alt' => '',
            'title' => '',
            'width' => NULL,
            'height' => NULL,
          ],
        ],
      ],
      'integer' => [
        'label' => 'Integer',
        'description' => '',
        'field_type' => 'integer',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'min' => NULL,
          'max' => NULL,
          'prefix' => '',
          'suffix' => '',
        ],
      ],
      'link' => [
        'label' => 'Link',
        'description' => '',
        'field_type' => 'link',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'title' => 1,
          'link_type' => LinkItemInterface::LINK_GENERIC,
        ],
      ],
      'list_float' => [
        'label' => 'List (float)',
        'description' => '',
        'field_type' => 'list_float',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'list_integer' => [
        'label' => 'List (integer)',
        'description' => '',
        'field_type' => 'list_integer',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'list_string' => [
        'label' => 'List (string)',
        'description' => '',
        'field_type' => 'list_string',
        'required' => FALSE,
        'translatable' => FALSE,
        'settings' => [],
        'default_value' => [],
        'default_value_callback' => '',
      ],
      'string' => [
        'label' => 'String value',
        'description' => '',
        'field_type' => 'string',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [],
      ],
      'string_long' => [
        'label' => 'string_long',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [],
        'field_type' => 'string_long',
      ],
      'telephone' => [
        'label' => 'Telephone',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [],
        'field_type' => 'telephone',
      ],
      'text' => [
        'label' => 'Text',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'allowed_formats' => [],
        ],
        'field_type' => 'text',
      ],
      'text_formatted' => [
        'label' => 'Formatted text',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          // @todo Add HTML formats.
          'allowed_formats' => [],
        ],
        'field_type' => 'text_l',
      ],
      'text_with_summary' => [
        'label' => 'Text with summary',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [
          'display_summary' => FALSE,
          'required_summary' => FALSE,
          'allowed_formats' => [],
        ],
        'field_type' => 'text_with_summary',
      ],
      'timestamp' => [
        'label' => 'Timestamp',
        'description' => '',
        'required' => FALSE,
        'translatable' => FALSE,
        'default_value' => [],
        'default_value_callback' => '',
        'settings' => [],
        'field_type' => 'timestamp',
      ],
    ],
    'field_storage' => [
      'boolean' => [
        'type' => 'boolean',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'comment' => [
        'type' => 'comment',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'date' => [
        'type' => 'datetime',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'daterange' => [
        'type' => 'daterange',
        'settings' => [
          'datetime_type' => 'datetime',
        ],
        'locked' => FALSE,
        'cardinality' => 1,
        'translatable' => TRUE,
        'indexes' => [],
        'persist_with_no_fields' => FALSE,
        'custom_storage' => FALSE,
      ],
      'decimal' => [
        'type' => 'decimal',
        'settings' => [
          'precision' => 10,
          'scale' => 2,
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'email' => [
        'type' => 'email',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'reference' => [
        'type' => 'entity_reference',
        'settings' => [
          'target_type' => 'node',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'file' => [
        'type' => 'file',
        'settings' => [
          'target_type' => 'file',
          'display_field' => FALSE,
          'display_default' => FALSE,
          'uri_scheme' => 'private',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'float' => [
        'type' => 'float',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'image' => [
        'type' => 'image',
        'settings' => [
          'target_type' => 'file',
          'display_field' => FALSE,
          'display_default' => FALSE,
          'uri_scheme' => 'public',
          'default_image' => [
            'alt' => '',
            'title' => '',
            'width' => NULL,
            'height' => NULL,
          ],
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'integer' => [
        'type' => 'integer',
        'settings' => [
          'unsigned' => FALSE,
          'size' => 'normal',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'link' => [
        'type' => 'link',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'list_float' => [
        'type' => 'list_float',
        'settings' => [
          'allowed_values' => [],
          'allowed_values_function' => '',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'list_integer' => [
        'type' => 'list_integer',
        'settings' => [
          'allowed_values' => [],
          'allowed_values_function' => '',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'list_string' => [
        'type' => 'list_string',
        'settings' => [
          'allowed_values' => [],
          'allowed_values_function' => '',
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'string' => [
        'type' => 'string',
        'cardinality' => 1,
        'custom_storage' => FALSE,
        'indexes' => [],
        'locked' => FALSE,
        'persist_with_no_fields' => FALSE,
        'settings' => [
          'max_length' => 255,
          'case_sensitive' => FALSE,
          'is_ascii' => FALSE,
        ],
        'translatable' => TRUE,
      ],
      'string_long' => [
        'type' => 'string_long',
        'settings' => [
          'case_sensitive' => FALSE,
        ],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'telephone' => [
        'type' => 'telephone',
        'settings' => [],
        'locked' => FALSE,
        'cardinality' => 1,
        'translatable' => TRUE,
        'indexes' => [],
        'persist_with_no_fields' => FALSE,
        'custom_storage' => FALSE,
      ],
      'text' => [
        'type' => 'text_long',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'text_formatted' => [
        'type' => 'text_long',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
      'timestamp' => [
        'type' => 'timestamp',
        'settings' => [],
        'cardinality' => 1,
        'translatable' => TRUE,
      ],
    ],
    'view_display' => [
      'boolean' => [
        'type' => 'boolean',
        'label' => 'above',
        'settings' => [
          'format' => 'default',
          'format_custom_FALSE' => '',
          'format_custom_TRUE' => '',
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'comment' => [
        'type' => 'comment_default',
        'label' => 'above',
        'settings' => [
          'view_mode' => 'default',
          'pager_id' => 0,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'date' => [
        'type' => 'datetime_default',
        'label' => 'above',
        'settings' => [
          'timezone_override' => '',
          'format_type' => 'medium',
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'daterange' => [
        'type' => 'daterange_default',
        'label' => 'above',
        'settings' => [
          'timezone_override' => '',
          'format_type' => 'medium',
          'from_to' => 'both',
          'separator' => '-',
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'decimal' => [
        'type' => 'number_decimal',
        'label' => 'above',
        'settings' => [
          'thousand_separator' => '',
          'decimal_separator' => '.',
          'scale' => 2,
          'prefix_suffix' => TRUE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'email' => [
        'type' => 'basic_string',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'reference' => [
        'type' => 'entity_reference_label',
        'label' => 'above',
        'settings' => [
          'link' => TRUE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'file' => [
        'type' => 'file_default',
        'label' => 'above',
        'settings' => [
          'use_description_as_link_text' => TRUE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'float' => [
        'type' => 'number_decimal',
        'label' => 'above',
        'settings' => [
          'thousand_separator' => '',
          'decimal_separator' => '.',
          'scale' => 2,
          'prefix_suffix' => TRUE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'image' => [
        'type' => 'image',
        'label' => 'above',
        'settings' => [
          'image_link' => '',
          'image_style' => '',
          'image_loading' => [
            'attribute' => 'lazy',
          ],
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'integer' => [
        'type' => 'number_integer',
        'label' => 'above',
        'settings' => [
          'thousand_separator' => '',
          'prefix_suffix' => TRUE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'link' => [
        'type' => 'link',
        'label' => 'above',
        'settings' => [
          'trim_length' => 80,
          'url_only' => FALSE,
          'url_plain' => FALSE,
          'rel' => '',
          'target' => '',
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'list_float' => [
        'type' => 'list_default',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'list_integer' => [
        'type' => 'list_default',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'list_string' => [
        'type' => 'list_default',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'string' => [
        'type' => 'string',
        'label' => 'inline',
        'settings' => [
          'link_to_entity' => FALSE,
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'string_long' => [
        'type' => 'basic_string',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'telephone' => [
        'type' => 'basic_string',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'text' => [
        'type' => 'text_default',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'text_formatted' => [
        'type' => 'text_default',
        'label' => 'above',
        'settings' => [],
        'third_party_settings' => [],
        'region' => 'content',
      ],
      'timestamp' => [
        'type' => 'timestamp',
        'label' => 'above',
        'settings' => [
          'date_format' => 'medium',
          'custom_date_format' => '',
          'timezone' => '',
          'tooltip' => [
            'date_format' => 'long',
            'custom_date_format' => '',
          ],
          'time_diff' => [
            'enabled' => FALSE,
            'future_format' => '@interval hence',
            'past_format' => '@interval ago',
            'granularity' => 2,
            'refresh' => 60,
          ],
        ],
        'third_party_settings' => [],
        'region' => 'content',
      ],
    ],
    'form_display' => [
      'boolean' => [
        'type' => 'boolean_checkbox',
        'region' => 'content',
        'settings' => [
          'display_label' => TRUE,
        ],
        'third_party_settings' => [],
      ],
      'comment' => [
        'type' => 'comment_default',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'date' => [
        'type' => 'datetime_default',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'daterange' => [
        'type' => 'daterange_default',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'decimal' => [
        'type' => 'number',
        'region' => 'content',
        'settings' => [
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'email' => [
        'type' => 'email_default',
        'region' => 'content',
        'settings' => [
          'placeholder' => '',
          'size' => 60,
        ],
        'third_party_settings' => [],
      ],
      'reference' => [
        'type' => 'entity_reference_autocomplete',
        'region' => 'content',
        'settings' => [
          'match_operator' => 'CONTAINS',
          'match_limit' => 10,
          'size' => 60,
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'file' => [
        'type' => 'file_generic',
        'region' => 'content',
        'settings' => [
          'progress_indicator' => 'throbber',
        ],
        'third_party_settings' => [],
      ],
      'float' => [
        'type' => 'number',
        'region' => 'content',
        'settings' => [
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'image' => [
        'type' => 'image_image',
        'region' => 'content',
        'settings' => [
          'progress_indicator' => 'throbber',
          'preview_image_style' => 'thumbnail',
        ],
        'third_party_settings' => [],
      ],
      'integer' => [
        'type' => 'number',
        'region' => 'content',
        'settings' => [
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'link' => [
        'type' => 'link_default',
        'region' => 'content',
        'settings' => [
          'placeholder_url' => '',
          'placeholder_title' => '',
        ],
        'third_party_settings' => [],
      ],
      'list_float' => [
        'type' => 'options_select',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'list_integer' => [
        'type' => 'options_select',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'list_string' => [
        'type' => 'options_select',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
      'string' => [
        'type' => 'string_textfield',
        'region' => 'content',
        'settings' => [
          'size' => 60,
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'string_long' => [
        'type' => 'string_textarea',
        'region' => 'content',
        'settings' => [
          'rows' => 5,
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'telephone' => [
        'type' => 'telephone_default',
        'region' => 'content',
        'settings' => [
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'text' => [
        'type' => 'text_textarea',
        'region' => 'content',
        'settings' => [
          'rows' => 5,
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'text_formatted' => [
        'type' => 'text_textarea',
        'region' => 'content',
        'settings' => [
          'rows' => 5,
          'placeholder' => '',
        ],
        'third_party_settings' => [],
      ],
      'timestamp' => [
        'type' => 'datetime_timestamp',
        'region' => 'content',
        'settings' => [],
        'third_party_settings' => [],
      ],
    ],
  ];

  /**
   * Constructs a DataManager object.
   *
   * The configuration parameters is expected to contain the external entity
   * type (key ExternalEntityTypeInterface::XNTT_TYPE_PROP), the field name
   * (key 'field_name') and the property name (key 'property_name') this data
   * processor will work on.
   *
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Drupal\file\FileRepository $file_repository
   *   The file repository.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   HTTP client.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache backend.
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   */
  public function __construct(
    TranslationInterface $string_translation,
    ConfigFactoryInterface $config_factory,
    LoggerChannelFactoryInterface $logger_factory,
    FileRepository $file_repository,
    FileSystemInterface $file_system,
    ClientInterface $http_client,
    EntityTypeManager $entity_type_manager,
    CacheBackendInterface $cache_backend,
    array $configuration = [],
  ) {
    $this->setStringTranslation($string_translation);
    $this->configFactory = $config_factory;
    $this->loggerChannelFactory = $logger_factory;
    $this->logger = $this->loggerChannelFactory->get('genoring_data_manager');
    $this->fileRepository = $file_repository;
    $this->fileSystem = $file_system;
    $this->httpClient = $http_client;
    $this->entityTypeManager = $entity_type_manager;
    $this->cache = $cache_backend;
    $configuration =
      $configuration
      ?: $this->configFactory->get(static::CONFIG_KEY)->get()
      ?: static::$defaultConfiguration;
    $this->setConfiguration($configuration);
    $this->dataLocators = [];
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
  ) {
    return new static(
      $container->get('string_translation'),
      $container->get('config.factory'),
      $container->get('logger.factory'),
      $container->get('file.repository'),
      $container->get('file_system'),
      $container->get('http_client'),
      $container->get('entity_type.manager'),
      $container->get('cache.default')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration;
    $this->dataLocators = [];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return static::$defaultConfiguration;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state,
  ) {
    $data_manager_id = $form['#attributes']['id'] ??= 'gr_dm';

    $form['info'] = [
      '#type' => 'item',
      '#markup' => $this->t(
        "Here, you can configure how and where your data are stored.",
      ),
    ];

    $form['actions'] ??= [
      '#type' => 'actions',
    ];

    $form = $this->buildDirectoryConfigurationForm($form, $form_state);

    // Data locator management.
    $form = $this->buildDataLocatorConfigurationForm($form, $form_state);

    return $form;
  }

  /**
   * Build the data locator configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The data locator configuration form.
   */
  public function buildDirectoryConfigurationForm(
    array $form,
    FormStateInterface $form_state,
  ) {
    $configuration = $this->getConfiguration();

    $form['data_uri'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Managed data URI'),
      '#description' => $this->t('Base directory URI in which all data files should be stored. The corresponding directory will contain managed data files, datasets and generated files. It is recommended to use the "private://" scheme to prevent unauthorized access to data files.'),
      '#attributes' => [
        'placeholder' => static::$defaultConfiguration['data_uri'],
      ],
      '#default_value' => $form_state->getValue('data_uri', $configuration['data_uri']),
      '#required' => TRUE,
    ];

    $form['upload_uri'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Upload URI'),
      '#description' => $this->t('Upload directory URI where file are uploaded before being managed. It is recommended to use the "private://" scheme to restrict access to uploaded files.'),
      '#attributes' => [
        'placeholder' => static::$defaultConfiguration['upload_uri'],
      ],
      '#default_value' => $form_state->getValue('upload_uri', $configuration['upload_uri']),
      '#required' => TRUE,
    ];

    $form['download_uri'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Download URI'),
      '#description' => $this->t('Download directory URI where files available for download will be placed. It can use a "public://" schema to allow public access.'),
      '#attributes' => [
        'placeholder' => static::$defaultConfiguration['download_uri'],
      ],
      '#default_value' => $form_state->getValue('download_uri', $configuration['download_uri']),
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * Build the data locator configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The data locator configuration form.
   */
  public function buildDataLocatorConfigurationForm(
    array $form,
    FormStateInterface $form_state,
  ) {
    $configuration = $this->getConfiguration();
    $data_manager_id = $form['#attributes']['id'] ??= 'gr_dm';

    $data_locator_manager = \Drupal::service('plugin.manager.genoring.data_locator');
    $data_locator_list_id = $data_manager_id . '_dls';
    $form['data_locators'] = [
      '#type' => 'details',
      '#title' => $this->t('Data locator settings'),
      '#description' => $this->t(
        'Data locator plugins are responsible for creating the appropriate directory structure to place data files according to their metadata.
Their order is important and directly impact the directory structure set: the first one is processed first and so on.
It is not advised to change this setting once data files have already been added to GenoRing.'
      ),
      '#tree' => TRUE,
      'expressions' => [
        '#type' => 'details',
        '#title' => $this->t('Help'),
        '#markup' => $this->t(
          '<h3>Data locator plugins</h3>
<p>Data locator plugins are used to automatically place files in the local file system according to their metadata.
The <i>root data locator</i> plugin is used to select the base file scheme (private/public) and base directory where all data files will be placed.
Then, the following data locator plugins will be evaluated in order from 1 to the end.
For each data locator plugin, the corresponding metadata rules (see below) will be evaluated against file metadata to determine if the data locator plugin should be used or not for that file.
Each line of the rules is a standalone rule. Those rules will be evaluated in order (from top to bottom) until a match is met. The first matching rule will determine if the data locator plugin should be used (&quot;+&quot; prefix) or ingnored(&quot;-&quot; prefix).
When a data locator plugin is used, it will alter the current file path provided by previous data locator plugins to add its own specificities.
For instance, a "Term locator" plugin can add a fixed subdirectory as well as a dynamic directory based on a metadata term (field) value.
Modifications are cumulated and that is why the order is important.</p>
<h3>Metadata rules</h3>
<p>Data locator plugins use expressions matched against file metadata to determine if a data locator plugin should be used or not.
Each expression line must follow the syntax:<br/>
&lt;+|-&gt;&lt;EXPRESSION&gt;[,&lt;EXPRESSION&gt;[,...]]
where:<ul>
  <li><strong>+|-</strong>: &quot;+&quot; or &quot;-&quot; are used respectively to involve or exclude the data locator plugin (only if the rest of the line matches the metadata).</li>
  <li><strong>EXPRESSION</strong>: &lt;TERM|JSON_PATH&gt;[=&lt;VALUE|\[COMA_SEPARATED_VALUES\]&gt;]</li>
  <li><strong>TERM</strong>: a metadata term (key) name.</li>
  <li><strong>JSON_PATH</strong>: a JSON path to locate the term in the metadata structure.</li>
  <li><strong>VALUE</strong>: a value to match.</li>
  <li><strong>COMA_SEPARATED_VALUES</strong>: a list of coma-separated values that must all be present in the array of values of the matched term.</li>
</ul>
Multiple expressions (coma-separated) on a single line must all match to have a line match.
Each line is considered separately, in order from the first one (top) to the last one, and the first line matching the metadata will determine if the data locator plugin will be used (line starting with &quot;+&quot;) or discarded (starting with &quot;-&quot;).
If no expression is provided, the data locator plugin will always be used. If expressions are specified but no line matches, the data locator plugin will not be used.
</p>
<p>
For example, to match a metadata that has a taxonomy value for a FASTA file or a GFF3 file, use the lines:
<pre>+taxonomy,format=fasta
+taxonomy,format=gff3
</pre>
But if you want to exclude the data locator plugin for files in which contributed "Jane" and "John", you can use the exclusion line:
<pre>-contributor=[Jane,John]
</pre>
Note: The above will not exclude file where only Jane or John contributed. If you wanted to exclude "Jane" OR "John", you would use instead:
<pre>-contributor=Jane
-contributor=John
</pre>
See documentation for more details and complex cases.</p>
'
        ),
      ],
    ];

    // Get current data locators selection.
    $data_locators =
      $form_state->get('data_locators')
      ?? $configuration['data_locators']
      ?? [];
    if (empty($data_locators)) {
      // Add an empty root data locator by default.
      $data_locators = [[]];
    }
    $form_state->set('data_locators', $data_locators);
    $num_data_locators = max(count($data_locators), 1);

    // Root data locator.
    $form['#parents'] ??= [];
    $form['data_locators']['root']['config'] = [
      '#type' => 'container',
      '#parents' => array_merge($form['#parents'], ['data_locators', 'root', 'config']),
    ];
    $data_manager_url = Url::fromRoute('genoring.data_manager_form');
    $form['data_locators']['root'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Root data locator'),
      '#description' => $this->t(
        'This data locator plugin is used to select the base directory where data files and directories will be placed. The base URI can be configured on the <a hrep="@data_manager_url">Data Manager page</a>.',
        ['@data_manager_url' => $data_manager_url->toString()]
      ),
      'base' => [
        '#type' => 'item',
        '#markup' => $this->t(
          'Base data URI: <code>@data_uri</code>',
          ['@data_uri' => $configuration['data_uri']]
        ),
      ],
    ];

    // Now add other data locators.
    $form['data_locators']['sublocators'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => $data_locator_list_id,
        'class' => ['genoring-data-locator-list'],
      ],
    ];
    for ($i = 1; $i < $num_data_locators; ++$i) {
      $form['data_locators']['sublocators']['dl_' . $i] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Data locator #@num', ['@num' => $i]),
        '#attributes' => ['class' => ['genoring-item'], 'data-weight' => $i - 1],
        '#prefix' => '<div class="draggable">',
        '#suffix' => '</div>',
        'weight' => [
          '#type' => 'weight',
          '#title' => $this->t('Order:'),
          '#default_value' => $i - 1,
          '#delta' => $num_data_locators,
          '#attributes' => ['class' => ['weight']],
          '#wrapper_attributes' => ['class' => ['container-inline']],
        ],
        'id' => [
          '#type' => 'select',
          '#title' => $this->t('Data locator:'),
          '#options' => $this->getDataLocatorOptions(),
          '#empty_option' => $this->t('- Select -'),
          '#default_value' => $data_locators[$i]['id'] ?? NULL,
          '#attributes' => [
            'data-genoring-action' => 'select',
            'data-genoring-locator' => $i,
            'autocomplete' => 'off',
          ],
          '#wrapper_attributes' => ['class' => ['container-inline']],
          '#validate' => [[static::class, 'validateDataLocatorAjax']],
          '#ajax' => [
            'callback' => [static::class, 'buildAjaxSublocatorsForm'],
            'wrapper' => $data_locator_list_id,
          ],
        ],
        'config' => [
          '#type' => 'container',
          '#parents' => array_merge($form['#parents'], ['data_locators', 'sublocators', 'dl_' . $i, 'config']),
        ],
        'rules' => [
          '#type' => 'textarea',
          '#title' => $this->t('Metadata rules:'),
          '#description' => $this->t('Metadata matching expression lines used to involve or exclude this data locator. See help above or module documentation for details.'),
          '#default_value' => $data_locators[$i]['rules'] ?? '',
        ],
        'notes' => [
          '#type' => 'textarea',
          '#title' => $this->t('Notes:'),
          '#default_value' => $data_locators[$i]['notes'] ?? '',
        ],
        'remove' => [
          '#type' => 'submit',
          // We must use a unique name to ensure the button can be identified as
          // the triggering element.
          '#name' => 'gr_dm_dls_rem_' . $i,
          '#value' => t('Remove'),
          '#attributes' => [
            'data-genoring-action' => 'remove',
            'data-genoring-locator' => $i,
          ],
          '#validate' => [[static::class, 'validateDataLocatorAjax']],
          '#ajax' => [
            'callback' => [static::class, 'buildAjaxSublocatorsForm'],
            'wrapper' => $data_locator_list_id,
          ],
        ],
      ];

      if (!empty($data_locators[$i]['id'])) {
        $locator = $data_locator_manager->createInstance(
          $data_locators[$i]['id'],
          $data_locators[$i]['config'] ?? []
        );
        $plugin_config_form_state = SubformState::createForSubform(
          $form['data_locators']['sublocators']['dl_' . $i]['config'],
          $form,
          $form_state
        );
        $form['data_locators']['sublocators']['dl_' . $i]['config'] =
          $locator->buildConfigurationForm(
            $form['data_locators']['sublocators']['dl_' . $i]['config'],
            $plugin_config_form_state
          );
        if (!isset($data_locators[$i]['rules'])) {
          // Provide default supported metadata as default.
          $user_input = $form_state->getUserInput();
          $user_input['data_locators']['sublocators']['dl_' . $i]['rules'] = implode("\n", $locator->getDefaultSupportedMetadata());
          $form_state->setUserInput($user_input);
        }
      }
    }

    $form['data_locators']['add_data_locator'] = [
      '#type' => 'submit',
      '#value' => t('Add data locator'),
      '#attributes' => [
        'data-genoring-action' => 'add',
      ],
      '#validate' => [[static::class, 'validateDataLocatorAjax']],
      '#ajax' => [
        'callback' => [static::class, 'buildAjaxSublocatorsForm'],
        'wrapper' => $data_locator_list_id,
      ],
    ];

    // Add JavaScript libraries.
    $form['#attached']['library'][] = 'core/drupal.ajax';
    $form['#attached']['library'][] = 'core/drupal.form';
    $form['#attached']['library'][] = 'genoring/genoring';

    return $form;
  }

  /**
   * Returns an array of available data locator plugins.
   *
   * @return array
   *   An array of options suitable for a form select element.
   */
  public function getDataLocatorOptions() :array {
    $data_locator_manager = \Drupal::service('plugin.manager.genoring.data_locator');
    $data_locator_options = [];
    $data_locators = $data_locator_manager->getDefinitions();
    foreach ($data_locators as $data_locator_id => $definition) {
      $data_locator_options[$data_locator_id] = $definition['label'];
    }
    return $data_locator_options;
  }

  /**
   * Returns the extra-data locator settings for AJAX calls.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The HTTP request object.
   *
   * @return array
   *   The 'sublocators' for part build by self::buildConfigurationForm().
   */
  public static function buildAjaxSublocatorsForm(
    array &$form,
    FormStateInterface &$form_state,
    Request $request,
  ) {
    return $form['data_locators']['sublocators'];
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
    $this->validateDirectoryConfigurationForm($form, $form_state);
    $this->validateDataLocatorConfigurationForm($form, $form_state);
  }

  /**
   * Validate directory configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function validateDirectoryConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
  }

  /**
   * Validate data locator configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function validateDataLocatorConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
  }

  /**
   * Validate data locator ajax actions.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public static function validateDataLocatorAjax(
    array &$form,
    FormStateInterface $form_state,
  ) {
    // Check for Ajax events.
    if ($trigger = $form_state->getTriggeringElement()) {
      $action = $trigger['#attributes']['data-genoring-action'] ?? '';
      $locator_number = $trigger['#attributes']['data-genoring-locator'] ?? NULL;
      // Add an empty root data locator by default.
      $data_locators = $form_state->get('data_locators') ?? [[]];
      if ('add' == $action) {
        $data_locators[] = [
          'id' => '',
          'notes' => '',
        ];
        $form_state->set('data_locators', $data_locators);
        $form_state->setRebuild(TRUE);
      }
      elseif (
        empty($locator_number)
        || (0 >= $locator_number)
        || (count($data_locators) <= $locator_number)
      ) {
        // Nothing to do: missing a valid locator number.
      }
      elseif ('remove' == $action) {
        $removed = array_splice($data_locators, $locator_number, 1);
        $data_locators_count = count($data_locators);
        $form_state->set('data_locators', $data_locators);
        // Update input order.
        $user_input = $form_state->getUserInput();
        for ($config_number = $locator_number; $config_number < $data_locators_count; ++$config_number) {
          // Shift form values.
          $form_state->setValueForElement(
            ['data_locators', 'sublocators', 'dl_' . $config_number],
            $data_locators[$config_number] ?? ''
          );
          $user_input['data_locators']['sublocators']['dl_' . $config_number] = $user_input['data_locators']['sublocators']['dl_' . ($config_number + 1)];
        }
        unset($user_input['data_locators']['sublocators']['dl_' . $data_locators_count]);
        $form_state->setUserInput($user_input);
        $form_state->setRebuild(TRUE);
      }
      elseif ('select' == $action) {
        $data_locators = $form_state->get('data_locators') ?? [];
        $data_locators[$locator_number]['id'] = $form_state->getValue($trigger['#array_parents'], '');
        $form_state->set('data_locators', $data_locators);
        $form_state->setRebuild(TRUE);
      }
    }

    // If rebuild needed, ignore validation.
    if ($form_state->isRebuilding()) {
      $form_state->clearErrors();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
    $this->submitDirectoryConfigurationForm($form, $form_state);
    $this->submitDataLocatorConfigurationForm($form, $form_state);
  }

  /**
   * Submit directory configuration from.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function submitDirectoryConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
    $old_configuration = $this->getConfiguration();
    $configuration = [
      'upload_uri' => $form_state->getValue(
        'upload_uri',
        $old_configuration['upload_uri']
      ),
      'data_uri' => $form_state->getValue(
        'data_uri',
        $old_configuration['data_uri']
      ),
      'download_uri' => $form_state->getValue(
        'download_uri',
        $old_configuration['download_uri']
      ),
      'data_locators' => [],
    ] + $old_configuration;
    $this->setConfiguration($configuration);
    $this->configFactory->getEditable(static::CONFIG_KEY)
      ->setData($configuration)
      ->save();
  }

  /**
   * Submit data locatro plugin configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function submitDataLocatorConfigurationForm(
    array &$form,
    FormStateInterface $form_state,
  ) {
    $configuration = $this->getConfiguration();
    $sub_locators = $form_state->getValue(['data_locators', 'sublocators'], []);
    // Only keep locators with a plugin set.
    $sub_locators = array_filter(
      array_values($sub_locators),
      function ($item) {
        return !empty($item['id']);
      }
    );
    // Reorder locators.
    uasort(
      $sub_locators,
      function ($a, $b) {
        return $a['weight'] <=> $b['weight'];
      }
    );
    // Only keep valid config elements.
    $sub_locators = array_map(
      function ($item) {
        return array_intersect_key(
          $item,
          [
            'id' => TRUE,
            'config' => TRUE,
            'rules' => TRUE,
            'notes' => TRUE,
          ]
        );
      },
      $sub_locators
    );
    // Trim rules.
    foreach ($sub_locators as &$sub_locator) {
      $sub_locator['rules'] = trim(preg_replace('/\s*\n\s*/', "\n", $sub_locator['rules'] ?? ''));
    }
    $configuration['data_locators'] = [
      [
        'id' =>
        $old_configuration['data_locators'][0]['id']
        ?? static::$defaultConfiguration['data_locators'][0]['id'],
        'config' =>
        $old_configuration['data_locators'][0]['config']
        ?? static::$defaultConfiguration['data_locators'][0]['config'],
        'rules' =>
        $old_configuration['data_locators'][0]['rules']
        ?? static::$defaultConfiguration['data_locators'][0]['rules'],
        'notes' =>
        $old_configuration['data_locators'][0]['notes']
        ?? static::$defaultConfiguration['data_locators'][0]['notes'],
      ],
      // phpcs:ignore
      ...$sub_locators
    ];
    $this->setConfiguration($configuration);
    $this->configFactory->getEditable(static::CONFIG_KEY)
      ->setData($configuration)
      ->save();
  }

  /**
   * Returns the ordered list of configured data locators.
   *
   * @return \Drupal\genoring\DataLocator\DataLocatorInterface[]
   *   The array of data locators.
   */
  public function getDataLocators() :array {
    if (empty($this->dataLocators)) {
      $this->dataLocators = [];
      $data_locator_manager = \Drupal::service('plugin.manager.genoring.data_locator');
      $configuration = $this->getConfiguration();
      foreach ($configuration['data_locators'] as $dl_index => $data_locator_settings) {
        try {
          $locator = $data_locator_manager->createInstance(
            $data_locator_settings['id'],
            $data_locator_settings['config'] ?? []
          );
          $this->dataLocators[$dl_index] = $locator;
        }
        catch (PluginNotFoundException $e) {
          $this->logger->error(
            'Failed to load data locator plugin #'
            . $dl_index
            . ' ('
            . $data_locator_settings['id']
            . '): '
            . $exception
          );
        }
      }
    }
    return $this->dataLocators;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareFileDirectory(string &$file_uri, bool $create_file = FALSE) :string {
    $real_path = '';
    if (!file_exists($file_uri)) {
      $directory = $this->fileSystem->dirname($file_uri);
      if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
        $this->logger->error('Failed to create directory @directory.', ['@directory' => $directory]);
        throw new DataManagerException('Failed to create directory ' . $directory . '.');
      }
      $real_path = $this->fileSystem->realpath($directory) . '/' . $this->fileSystem->basename($file_uri);
    }
    if ($create_file) {
      $file = $this->fileRepository->writeData('', $file_uri, FileExists::Rename);
      $file_uri = $file->getFileUri();
      $real_path = $this->fileSystem->realpath($file_uri);
    }

    return $real_path;
  }

  /**
   * {@inheritdoc}
   */
  public function generateSymlink(string $target_uri, string $link_uri) :?FileInterface {
    $target_path = $this->fileSystem->realpath($target_uri);
    $link_name = $this->fileSystem->basename($link_uri);
    $link_directory = $this->fileSystem->dirname($link_uri);
    if (!$this->fileSystem->prepareDirectory($link_directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger->error('Failed to create directory @directory.', ['@directory' => $link_directory]);
      throw new DataManagerException('Failed to create directory ' . $link_directory . '.');
    }
    $link_path = $this->fileSystem->realpath($link_directory) . '/' . $link_name;
    if ($target_path != $link_path) {
      if (is_link($link_path)) {
        // Remove previous symlink to update.
        $this->logger->warning('The symlink "@symlink" already exists and will be replaced.', ['@symlink' => $link_path]);
        unlink($link_path);
      }
      // Generate link if possible.
      if (file_exists($link_path)) {
        $this->logger->warning('Unable to create symlink "@symlink": a file with the same name already exists.', ['@symlink' => $link_path]);
      }
      elseif (symlink($this->getRelativePath($link_path, $target_path), $link_path)) {
        $file_metadata['references'] = [$uri];
        // Manage link in Drupal.
        $file_storage = $this->entityTypeManager->getStorage('file');
        $file_query = $file_storage->getQuery();
        $file_ids = $file_query->accessCheck(TRUE)->condition('uri', $link_path)->execute();
        if (empty($file_ids)) {
          $file = File::create([
            'filename' => $this->fileSystem->basename($link_path),
            'uri' => $link_uri,
            'status' => File::STATUS_PERMANENT,
            'uid' => \Drupal::currentUser()->id(),
          ]);
          $file->save();
        }
      }
      else {
        $this->logger->warning(
          'Unable to create symlink "@symlink" for target "@target" (relative path: "@relative"): @error.',
          [
            '@symlink' => $link_path,
            '@target' => $target_path,
            '@relative' => $this->getRelativePath($link_path, $target_path),
            '@error' => error_get_last()['message'] ?? '',
          ]
        );
      }
    }
    return $file ?? NULL;
  }

  /**
   * Builds a relative path from a path to another.
   *
   * @param string $from
   *   Source path.
   * @param string $to
   *   Target path.
   *
   * @return string
   *   The relative path.
   */
  protected function getRelativePath(string $from, string $to) :string {
    $from = rtrim(dirname($from), '/');
    $to = rtrim($to, '/');
    $from_parts = explode('/', $from);
    $to_parts = explode('/', $to);
    $common_length = 0;
    $min_length = min(count($from_parts), count($to_parts));
    for ($i = 0; $i < $min_length; ++$i) {
      if ($from_parts[$i] === $to_parts[$i]) {
        $common_length = $i + 1;
      }
      else {
        break;
      }
    }
    $relative_parts = array_fill(0, count($from_parts) - $common_length, '..');
    $relative_parts = array_merge($relative_parts, array_slice($to_parts, $common_length));
    return implode('/', $relative_parts);
  }

  /**
   * {@inheritdoc}
   */
  public function getFile(array $metadata) :?FileInterface {
    $file = NULL;
    $file_uri = $this->locateUri($metadata, 'file');
    if (!empty($file_uri)) {
      $file_storage = $this->entityTypeManager->getStorage('file');
      $file_query = $file_storage->getQuery();
      $file_ids = $file_query->accessCheck(TRUE)->condition('uri', $file_uri)->execute();
      if (!empty($file_ids)) {
        $file = $file_storage->load(current($file_ids));
      }
    }
    return $file;
  }

  /**
   * {@inheritdoc}
   */
  public function saveGenoringInput(array $input_value) :array {
    $files = [];
    $uris = $input_value['uris'] ?? [];
    $metadata = $input_value['metadata'] ?? [];
    if (!empty($input_value['file_type'])) {
      $metadata['format'] = $input_value['file_type'];
    }
    $upload_base = $this->getConfiguration()['upload_uri'] . '/';
    $file_storage = $this->entityTypeManager->getStorage('file');
    foreach ($uris as $uri) {
      $file_metadata = $metadata;
      // Check if file is already managed.
      $file_query = $file_storage->getQuery();
      $file_ids = $file_query->accessCheck(TRUE)->condition('uri', $uri)->execute();
      if (!empty($file_ids)) {
        // Managed file.
        $file = $file_storage->load(current($file_ids));
      }
      if (empty($file)) {
        // Either the file is not managed or failed to load, save a new one.
        $file_metadata['source'] = [$uri];
        // Manage URLs.
        if (0 === strpos($uri, 'http')) {
          // Download file to upload directory.
          try {
            $response = $this->httpClient->get($uri);
            $file_content = (string) $response->getBody();
            $basename = $this->fileSystem->basename(parse_url($uri, PHP_URL_PATH));
            // Add extension if missing.
            if (FALSE === strrpos($basename, '.') && !empty($metadata['format'])) {
              if (is_array($metadata['format'])) {
                $basename .= '.' . reset($metadata['format']);
              }
              elseif (is_string($metadata['format'])) {
                $basename .= '.' . $metadata['format'];
              }
            }
            $upload_uri = $upload_base . $basename;
            $uri = $this->fileSystem->saveData($file_content, $upload_uri, FileExists::Rename);
            \Drupal::messenger()->addMessage(
              t(
                'Successfully downloaded file @url as @basename.',
                [
                  '@url' => $uri,
                  '@basename' => $basename,
                ]
              )
            );
          }
          catch (\Exception $exception) {
            $uri = '';
            \Drupal::logger('genoring')->error('Failed to download file ' . $uri . ': ' . $exception->getMessage());
          }
        }
        if (file_exists($uri)) {
          // Save it to a new managed file.
          $file = File::create([
            'filename' => $this->fileSystem->basename($uri),
            'uri' => $uri,
            'status' => File::STATUS_PERMANENT,
            'uid' => \Drupal::currentUser()->id(),
          ]);
          $file->save();
        }
      }

      // Now work on the managed file.
      if (!empty($file)) {
        // Check if in upload directory and needs to be moved.
        if (0 === strpos($uri, $upload_base)) {
          // Move file to source directory and set permanent.
          $file_metadata['isSource'] = TRUE;
          $file = $this->placeFile($file, $file_metadata);
        }
        else {
          // Check if managed by GenoRing.
          $current_metadata = $this->getMetadata($file);
          if (!empty($current_metadata)) {
            // Already managed by GenoRing, link it.
            $expected_uri = $data_manager->locateUri($file_metadata);
            $link_file = $this->generateSymlink($uri, $expected_uri);
            if (isset($link_file)) {
              $file = $link_file;
            }
            // @todo Should we merge original metadata?
            $this->setMetadata($expected_uri, $file_metadata);
            // @todo Update linked file metadata.
          }
          else {
            // Not managed by GenoRing yet, move it and manage it.
            $file_metadata['isSource'] = TRUE;
            $file = $this->placeFile($file, $file_metadata);
          }
        }
        $files[] = $file;

        // Process file if needed.
        if (!empty($input_value['processing']) && !empty($input_value['file_type'])) {
          $operations = $this->getPreProcessingOperations($input_value['file_type']);
        }
        if (!empty($operations)) {
          foreach ($operations as $operation_id => $operation) {
            // @todo Check if we got a subform for the operation.
            // Create sub-from and sub-form state.
            // Set 'uri' state.
            // Call submit handler $operation['submit'].
            // Check if URI changed and update if needed.
            // Save processed file metadata.
          }
        }
      }
    }
    return $files;
  }

  /**
   * {@inheritdoc}
   */
  public function locateUri(
    array $metadata,
    bool $provide_default = FALSE,
    ?array &$backtrace = NULL,
  ) :string {
    $configuration = $this->getConfiguration();
    $data_locators = $this->getDataLocators();
    if (3 == func_num_args()) {
      $backtrace = [];
    }
    try {
      if (empty(trim($metadata['type'] ?? ''))) {
        throw new MetadataException('Invalid metadata: missing "type" field with a non-empty value.');
      }
      if (('file' == $metadata['type'])
          && (!isset($metadata['filename']) || ('' == trim($metadata['filename'])))
      ) {
        throw new MetadataException('Invalid metadata: missing "filename" field with a non-empty value.');
      }
      // Get base.
      $default_locator = $data_locators[0];
      unset($data_locators[0]);
      $uri = $base_uri = $default_locator->locateUri($metadata, $metadata['filename']);
      if (isset($backtrace)) {
        $backtrace[0] = ['uri' => $uri];
      }
      foreach ($data_locators as $dl_index => $data_locator) {
        $dl_settings = $configuration['data_locators'][$dl_index];
        if (isset($backtrace)) {
          $rules_backtrace = [];
        }
        if (empty($dl_settings['rules'])
            || $this->matchMetadataAny($dl_settings['rules'], $metadata, $rules_backtrace)
        ) {
          $uri = $data_locator->locateUri($metadata, $uri);
          if (isset($backtrace)) {
            $backtrace[$dl_index] = [
              'uri' => $uri,
              'rules' => $rules_backtrace,
            ];
          }
        }
        elseif (isset($backtrace)) {
          $backtrace[$dl_index] = [
            'excluded' => TRUE,
            'rules' => $rules_backtrace,
          ];
        }
      }
      // Check if we should return an empty string.
      if (!$provide_default && ($base_uri == $uri)) {
        $uri = '';
      }
    }
    catch (MetadataException $exception) {
      if (isset($backtrace)) {
        // Rethrow.
        throw $exception;
      }
      $this->logger->error('::locateUri(): ' . $exception);
      $uri = '';
    }
    return $uri;
  }

  /**
   * {@inheritdoc}
   */
  public function placeFile(
    FileInterface $file,
    array $metadata,
  ) :FileInterface {
    // Check if we got metadata.
    $empty_metadata = FALSE;
    if (empty($metadata)) {
      $empty_metadata = TRUE;
    }
    // Compute location according to metadata.
    $metadata['type'] = 'file';
    $metadata['filename'] ??= $file->getFilename();
    $metadata['uri'] ??= $file->getFileUri();
    // Check if file uses a different name from its original one.
    if ($metadata['uri'] != $file->getFileUri()) {
      // Keep track of previous file name.
      if (empty($metadata['source'])) {
        $metadata['source'] = [$metadata['uri']];
      }
      elseif (is_array($metadata['source'])) {
        $metadata['source'][] = $metadata['uri'];
      }
      else {
        $metadata['source'] = [$metadata['source'], $metadata['uri']];
      }
    }
    $location = $this->locateUri($metadata, 'file');
    if (empty($location)) {
      throw new DataManagerException('Failed to get a URI to place file.');
    }
    // Make sure the file needs to be linked or moved.
    if ($location != $file->getFileUri()) {
      // Complete current metadata with the existing one.
      $current_metadata = $this->getMetadata($file);
      $metadata = $metadata + $current_metadata;
      // Check if managed by GenoRing.
      if (empty($current_metadata)) {
        // Not managed, move.
        // Keep track of previous locations.
        if (empty($metadata['source'])) {
          $metadata['source'] = [$file->getFileUri()];
        }
        elseif (is_array($metadata['source'])) {
          $metadata['source'][] = $file->getFileUri();
        }
        else {
          $metadata['source'] = [$metadata['source'], $file->getFileUri()];
        }
        // Prepare directory.
        $this->prepareFileDirectory($location);
        // Move file.
        $file = $this->fileRepository->move(
          $file,
          $location,
          FileExists::Rename
        );
        $file->setPermanent();
        $file->save();
        $metadata['filename'] = $file->getFilename();
        $metadata['uri'] = $file->getFileUri();
      }
      else {
        // Managed, link.
        $link_file = $this->generateSymlink($file->getFileUri(), $location);
        if (isset($link_file)) {
          $file = $link_file;
          // @todo Update linked file metadata.
        }
      }
    }
    elseif (!$file->isPermanent()) {
      $file->setPermanent();
      $file->save();
    }

    // Update metadata.
    $this->setMetadata($file, $metadata);
    return $file;
  }

  /**
   * {@inheritdoc}
   */
  public function placeFiles(array|string $file_list) :array {
    throw new DataManagerException('Not implemented');
  }

  /**
   * {@inheritdoc}
   */
  public function placePackage(
    string $zip_uri,
    array|string $metadata,
    bool $extract_files = FALSE,
  ) :array {
    throw new DataManagerException('Not implemented');
  }

  /**
   * {@inheritdoc}
   */
  public function listFiles(string $directory, string|array $extensions = [], int $flags = 0) :array {
    $file_list = [];
    if (empty($directory)) {
      // Missing a directory.
      return $file_list;
    }
    if ('/' !== $directory[-1]) {
      // Make sure we end by a slash.
      $directory .= '/';
    }
    $include_subdir = ($flags & static::LIST_SUBDIRS);
    $include_hidden = ($flags & static::LIST_HIDDEN);
    $include_links = ($flags & static::LIST_LINKS);
    if (!empty($extensions)) {
      if (is_array($extensions)) {
        $extension_regex = implode('|', $extensions);
      }
      else {
        $extension_regex = $extensions;
      }
      $extension_regex = str_replace('/', '', $extension_regex);
      $extension_regex = '/^(?:' . $extension_regex . ')$/';
      if ($flags & static::LIST_EXT_CASE_INSENSITIVE) {
        $extension_regex .= 'i';
      }
      $invert = FALSE;
      if ($flags & static::LIST_EXT_INVERT_MATCH) {
        $invert = TRUE;
      }
    }
    if ($path = $this->fileSystem->realpath($directory)) {
      $iterator = new \DirectoryIterator($path);
      foreach ($iterator as $file_info) {
        // Filter extensions.
        if ($file_info->isFile()) {
          $filename = $file_info->getFilename();
          if ((empty($extension_regex)
                || (preg_match($extension_regex, $file_info->getExtension()) xor $invert))
              && (($filename[0] !== '.') || $include_hidden)
              && (!$file_info->isLink() || $include_links)
          ) {
            $file_list[$directory . $filename] = $filename . ($file_info->isLink() ? '*' : '');
          }
        }
        elseif ($include_subdir
          && $file_info->isDir()
          && !$file_info->isDot()
          && (($file_info->getFilename()[0] !== '.') || $include_hidden)
          && (!$file_info->isLink() || $include_links)
        ) {
          // List sub-directories.
          $file_list += $this->listFiles($directory . $file_info->getFilename(), $extensions, $flags);
        }
      }
    }
    return $file_list;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetadata(string|FileInterface $file) :array {
    $metadata = [];
    $meta_metadata = [
      'type' => 'file',
      'filename' => 'metadata',
      'format' => 'metadata',
      'targetType' => 'file',
      'targetUri' => is_string($file) ? $file : $file->getFileUri(),
    ];
    $metadata_uri = $this->locateUri($meta_metadata, 'file');
    if (!empty($metadata_uri) && file_exists($metadata_uri)) {
      try {
        $metadata = Yaml::parseFile($metadata_uri);
      }
      catch (ParseException $e) {
        $err_line = $e->getParsedLine();
        if (0 <= $err_line) {
          $this->logger->error(
            'Failed to parse Yaml file "'
            . $metadata_uri
            . '" near line '
            . $e->getParsedLine()
            . ': '
            . $e->getSnippet()
          );
        }
        else {
          $this->logger->error(
            'Failed to parse Yaml file "'
            . $metadata_uri
            . '"'
          );
        }
      }
    }
    return $metadata;
  }

  /**
   * {@inheritdoc}
   */
  public function setMetadata(
    string|FileInterface $file,
    array $metadata,
    int $mode = DataManagerInterface::METADATA_REPLACE,
  ) :DataManagerInterface {
    $meta_metadata = [
      'type' => 'file',
      'filename' => 'metadata',
      'format' => 'metadata',
      'targetType' => 'file',
      'targetUri' => is_string($file) ? $file : $file->getFileUri(),
    ];
    $metadata_uri = $this->locateUri($meta_metadata, 'file');

    if (!empty($metadata_uri)) {
      $old_metadata = [];
      if (file_exists($metadata_uri)) {
        if (static::METADATA_COMPLETE == $mode) {
          $old_metadata = $this->getMetadata($file);
          $metadata = NestedArray::mergeDeep(
            $metadata,
            $old_metadata
          );
        }
        elseif (static::METADATA_OVERRIDE == $mode) {
          $old_metadata = $this->getMetadata($file);
          $metadata = NestedArray::mergeDeep(
            $old_metadata,
            $metadata
          );
        }
        elseif (static::METADATA_REMOVE == $mode) {
          // Remove existing metadata file.
          $file_storage = $this->entityTypeManager->getStorage('file');
          $file_query = $file_storage->getQuery();
          $file_ids = $file_query->accessCheck(TRUE)->condition('uri', $metadata_uri)->execute();
          if (empty($file_ids)) {
            // Not managed.
            unlink($metadata_uri);
          }
          else {
            // Managed.
            $metadata_file = $file_storage->load(current($file_ids));
            $metadata_file->delete();
          }
          return $this;
        }
      }
      // Prepare directory.
      $directory = $this->fileSystem->dirname($metadata_uri);
      if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
        $this->logger->error('Failed to create directory @directory.', ['@directory' => $directory]);
      }
      else {
        $metadata_yaml = Yaml::dump($metadata);
        $this->fileRepository->writeData($metadata_yaml, $metadata_uri, FileExists::Replace);
      }
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function matchMetadata(
    string $metadata_expression,
    array $metadata,
  ) :bool {
    // Separate expressions.
    $expressions = preg_split('/(?<!\\\\),(?![^\\[]*\\])/', $metadata_expression);
    $match = TRUE;
    foreach ($expressions as $expression) {
      if (1 > strlen($expression)) {
        continue 1;
      }
      elseif ('=' === $expression[0]) {
        // Invalid syntax.
        $match = FALSE;
        break 1;
      }
      // Separate optional value specification.
      @[$expression, $expected_value] = preg_split('/=/', $expression, 2);
      // Check if JSON Path or just term name.
      if ('$' === $expression[0]) {
        // JSON Path.
        try {
          $json_object = new JsonObject($metadata, TRUE);
          $value = $json_object->get($expression);
          // Since we are using "smart get", JsonObject returns FALSE when no
          // match.
          if (FALSE === $value) {
            $match = FALSE;
            break 1;
          }
        }
        catch (InvalidJsonException | InvalidJsonPathException $exception) {
          $this->logger->error(
            'Failed to parse JSON Path ('
            . $configuration['term']
            . '): '
            . $exception
          );
        }

      }
      else {
        // Check if term is there.
        if (!array_key_exists($expression, $metadata)) {
          // Term not in metadata.
          $match = FALSE;
          break 1;
        }
        $value = $metadata[$expression];
      }
      // Check if it should contain a specific value.
      if (isset($expected_value)) {
        // Yes check the value.
        if (('[' == $expected_value[0]) && (']' == $expected_value[-1])) {
          // It should be an array, check if it is actually an array.
          if (!is_array($value)) {
            // Expecting an array but got a scalar.
            $match = FALSE;
            break 1;
          }
          // Get the list of values to check.
          $expected_values = preg_split('/(?<!\\\\),/', substr($expected_value, 1, -1));
          foreach ($expected_values as $expected_value) {
            if (1 > strlen($expected_value)) {
              continue 1;
            }
            if (!in_array($expected_value, $value)) {
              // At least an expected value is missing.
              $match = FALSE;
              break 2;
            }
          }
        }
        else {
          // Check if the expected value is there...
          if (is_array($value)) {
            if (!in_array($expected_value, $value)) {
              $match = FALSE;
              break 1;
            }
          }
          elseif ($value != $expected_value) {
            $match = FALSE;
            break 1;
          }
        }
      }
      else {
        // No expected value but something must be set.
        if (!isset($value) || ('' === $value)) {
          $match = FALSE;
          break 1;
        }
      }
    }
    return $match;
  }

  /**
   * {@inheritdoc}
   */
  public function matchMetadataAny(
    string|array $metadata_expressions,
    array $metadata,
    ?array &$rules_backtrace = NULL,
  ) :bool {
    $match = FALSE;
    if (3 == func_num_args()) {
      $rules_backtrace = [];
    }
    if (is_string($metadata_expressions)) {
      $metadata_expressions = explode("\n", $metadata_expressions);
    }
    $exclude_rules_count = 0;
    foreach ($metadata_expressions as $metadata_expression) {
      // Check for inclusion/exclusion.
      $include = TRUE;
      if ('-' == $metadata_expression[0]) {
        $include = FALSE;
        $metadata_expression = substr($metadata_expression, 1);
      }
      elseif ('+' == $metadata_expression[0]) {
        $metadata_expression = substr($metadata_expression, 1);
      }
      // Check if line matches.
      if ($this->matchMetadata($metadata_expression, $metadata)) {
        if ($include) {
          $match = TRUE;
        }
        if (isset($rules_backtrace)) {
          $rules_backtrace[$metadata_expression] = $include ? 'include' : 'exclude';
        }
        break;
      }
      else {
        if (!$include) {
          ++$exclude_rules_count;
        }
        if (isset($rules_backtrace)) {
          $rules_backtrace[$metadata_expression] = 'not matched (no ' . ($include ? 'include' : 'exclude') . ')';
        }
      }
    }
    // Check if we only had exclusion rules that all failed.
    if (!$match && ($exclude_rules_count == count($metadata_expressions))) {
      if (isset($rules_backtrace)) {
        $rules_backtrace['only non-matching exclusion rules'] = 'include';
      }
      $match = TRUE;
    }
    return $match;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataModelDefinitions() :array {
    if ($cache = $this->cache->get(static::CACHE_KEY_DATA_MODELS)) {
      $data_model_definitions = $cache->data;
    }
    else {
      $event = new DataModelsEvent();
      \Drupal::service('event_dispatcher')->dispatch(
        $event,
        GenoringEvents::DATA_MODELS
      );
      $data_model_definitions = $event->getDataModelDefinitions();
      $this->cache->set(static::CACHE_KEY_DATA_MODELS, $data_model_definitions, CacheBackendInterface::CACHE_PERMANENT);
    }
    return $data_model_definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getFileTypes() :array {
    if ($cache = $this->cache->get(static::CACHE_KEY_FILE_TYPES)) {
      $file_types = $cache->data;
    }
    else {
      $event = new FileTypesEvent();
      \Drupal::service('event_dispatcher')->dispatch(
        $event,
        GenoringEvents::FILE_TYPES
      );
      $file_types = $event->getFileTypes();
      $this->cache->set(static::CACHE_KEY_FILE_TYPES, $file_types, CacheBackendInterface::CACHE_PERMANENT);
    }
    return $file_types;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetadataFields(string $type) :array {
    if ($cache = $this->cache->get(static::CACHE_KEY_TYPE_METADATA)) {
      $metadata_fields = $cache->data;
    }
    else {
      $event = new DataTypeMetadataEvent();
      \Drupal::service('event_dispatcher')->dispatch(
        $event,
        GenoringEvents::DATA_TYPE_METADATA
      );
      $metadata_fields = $event->getMetadataFields();
      $this->cache->set(static::CACHE_KEY_TYPE_METADATA, $metadata_fields, CacheBackendInterface::CACHE_PERMANENT);
    }
    return $metadata_fields[$type] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function getPreProcessingOperations(string $type) :array {
    if ($cache = $this->cache->get(static::CACHE_KEY_TYPE_OPERATIONS)) {
      $operations = $cache->data;
    }
    else {
      $event = new DataTypeOperationsEvent();
      \Drupal::service('event_dispatcher')->dispatch(
        $event,
        GenoringEvents::DATA_TYPE_OPERATIONS
      );
      $operations = $event->getPreProcessingOperations();
      $this->cache->set(static::CACHE_KEY_TYPE_OPERATIONS, $operations, CacheBackendInterface::CACHE_PERMANENT);
    }
    return $operations[$type] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function generateFormField(
    string $field_name,
    array $field_info,
  ) :array {
    $element = $this->dataFieldDefinitions['form_element'][$field_info['type']] ?? [];
    if (empty($element['#type'])) {
      $element['#type'] = $field_info['type'];
    }
    if (!empty($field_info['label'])) {
      $element['#title'] = $field_info['label'];
    }
    $element = NestedArray::mergeDeep(
      $element,
      $field_info['overrides']['form_element'] ?? []
    );
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function generateFieldConfig(
    string $field_name,
    array $field_info,
  ) :array {
    if (empty($field_info['type'])) {
      $this->logger->warning(
        'Invalid data model model_fields definition: '
        . print_r($field_info, TRUE)
      );
      $field_info['type'] = 'string';
    }
    $base_config = $this->dataFieldDefinitions['field_config'][$field_info['type']] ?? [];
    if (!empty($field_info['label'])) {
      $base_config['label'] = $field_info['label'];
    }
    if (!empty($field_info['description'])) {
      $base_config['description'] = $field_info['description'];
    }
    $element = NestedArray::mergeDeep(
      $base_config,
      $field_info['overrides']['field_config'] ?? []
    );
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function generateFieldStorage(
    string $field_name,
    array $field_info,
  ) :array {
    $element = NestedArray::mergeDeep(
      $this->dataFieldDefinitions['field_storage'][$field_info['type']] ?? [],
      $field_info['overrides']['field_storage'] ?? []
    );
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function generateFieldViewDisplay(
    string $field_name,
    array $field_info,
  ) :array {
    $element = NestedArray::mergeDeep(
      $this->dataFieldDefinitions['view_display'][$field_info['type']] ?? [],
      $field_info['overrides']['view_display'] ?? []
    );
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function generateFieldFormDisplay(
    string $field_name,
    array $field_info,
  ) :array {
    $element = NestedArray::mergeDeep(
      $this->dataFieldDefinitions['form_display'][$field_info['type']] ?? [],
      $field_info['overrides']['form_display'] ?? []
    );
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function generateFieldMapping(
    string $field_name,
    string|array $data_processors = [],
    array $overrides = [],
  ) :array {
    if (is_string($data_processors)) {
      if ('' === $data_processors) {
        $data_processors = [];
      }
      elseif ('default' == $data_processors) {
        $data_processors = [
          [
            'id' => 'default',
            'config' => [],
          ],
        ];
      }
      else {
        // Not supported.
        // @todo Warn.
        $data_processors = [
          [
            'id' => $data_processors,
            'config' => [],
          ],
        ];
      }
    }
    $field_mapping = NestedArray::mergeDeep(
      [
        'id' => 'generic',
        'config' => [
          'property_mappings' => [
            'value' => [
              'id' => 'direct',
              'config' => [
                'mapping' => $field_name,
                'required_field' => TRUE,
                'main_property' => TRUE,
                'data_processors' => $data_processors,
                'description' => '',
              ],
            ],
          ],
        ],
      ],
      $overrides
    );
    return $field_mapping;
  }

  /**
   * Generates a complete storage client config with the given overrides.
   *
   * @param string $plugin_id
   *   The storage client plugin id to use.
   * @param array $storage_config_override
   *   Storage client config overrides.
   * @param array $aggregation_config
   *   Storage client aggregation settings used by the group aggregator.
   *
   * @return array
   *   An array containing the plugin identifier, the storage client config, and
   *   the aggregation settings.
   */
  public function generateStorageClientConfig(
    string $plugin_id,
    array $storage_config_override = [],
    array $aggregation_config = [],
  ) :array {
    // Storage client config.
    $tsv_storage_client = \Drupal::service('plugin.manager.external_entities.storage_client')->createInstance($plugin_id, []);
    $storage_config = NestedArray::mergeDeep(
      $tsv_storage_client->defaultConfiguration(),
      $storage_config_override
    );
    // Aggregation config.
    $aggregation_config = NestedArray::mergeDeep(
      [
        'groups' => [],
        'group_prefix_strip' => FALSE,
        'id_field' => '',
        'mode' => GroupAggregator::STORAGE_CLIENT_MODE_READONLY,
      ],
      $aggregation_config
    );
    return [
      'id' => $plugin_id,
      'config' => $storage_config,
      'aggr' => $aggregation_config,
    ];
  }

}
