<?php

namespace Drupal\genoring\DataManager;

use Drupal\file\FileInterface;

/**
 * Interface for data manager.
 *
 * A data manager is used to locate data, manage uploaded files, and manage
 * datasets with their fields and associated metadata.
 *
 * Metadata is the data related to a given data element, usualy a file or a
 * directory (dataset). The metadata is structured as an associative array with
 * multiple levels. Keys are the metdata element names and values can be of any
 * type (single string, multiple strings, numeric values, a boolean,
 * sub-arrays, ...).
 * Most metadata are defined by the Dublin Core project and some other are
 * defined within the GenoRing project. Some typical metadata keys used by
 * GenoRing are:
 *
 * @todo To re-order and adjust (tag if from Dublin core or other, add expected
 * structure description, examples, etc.).
 *
 * - identifier: An unambiguous reference to the resource within a given context.
 *     Recommended practice is to identify the resource by means of a string conforming to an identification system. Examples include International Standard Book Number (ISBN), Digital Object Identifier (DOI), and Uniform Resource Name (URN). Persistent identifiers should be provided as HTTP URIs.
 *     Dataset URL or file URI.
 * - title: A name given to the resource.
 * - type: The nature or genre of the resource.
 *     Recommended practice is to use a controlled vocabulary such as the DCMI Type Vocabulary [DCMI-TYPE]. To describe the file format, physical medium, or dimensions of the resource, use the property Format.
 *     Typical values are "file", "directory", or "url".
 * - format: The file format, physical medium, or dimensions of the resource.
 *     Recommended practice is to use a controlled vocabulary where available. For example, for file formats one could use the list of Internet Media Types [MIME]. Examples of dimensions include size and duration.
 *     Used for file format and MIME types as synonyms (ex. 'fasta', and 'text/x-fasta' can be used as a synonym). Also used to identify 'metadata' file types. See https://www.iana.org/assignments/media-types/media-types.xhtml
 * - hasFormat: [] A related resource that is substantially the same as the pre-existing described resource, but in another format.
 *     This property is intended to be used with non-literal values. This property is an inverse property of Is Format Of.
 *     Filled by file converters on the source metadata.
 * - isFormatOf: A pre-existing related resource that is substantially the same as the described resource, but in another format.
 *     This property is intended to be used with non-literal values. This property is an inverse property of Has Format.
 *     Filled by file converters on the new generated file metadata.
 * - hasPart: [] A related resource that is included either physically or logically in the described resource.
 *     This property is intended to be used with non-literal values. This property is an inverse property of Is Part Of.
 *     Used by datasets to list included files.
 * - isPartOf: [] A related resource in which the described resource is physically or logically included.
 *     This property is intended to be used with non-literal values. This property is an inverse property of Has Part.
 *     Used by files that are part of datasets.
 * - coverage: [] The spatial or temporal topic of the resource, spatial applicability of the resource, or jurisdiction under which the resource is relevant.
 *     Spatial topic and spatial applicability may be a named place or a location specified by its geographic coordinates. Temporal topic may be a named period, date, or date range. A jurisdiction may be a named administrative entity or a geographic place to which the resource applies. Recommended practice is to use a controlled vocabulary such as the Getty Thesaurus of Geographic Names [TGN]. Where appropriate, named places or time periods may be used in preference to numeric identifiers such as sets of coordinates or date ranges. Because coverage is so broadly defined, it is preferable to use the more specific subproperties Temporal Coverage and Spatial Coverage.
 *     Used for file content (ex. "sequences", "alignment", descriptor number or name, etc.).
 * - created: Date of creation of the resource.
 *     Recommended practice is to describe the date, date/time, or period of time as recommended for the property Date, of which this is a subproperty.
 * - modified: Date on which the resource was changed.
 *     Recommended practice is to describe the date, date/time, or period of time as recommended for the property Date, of which this is a subproperty.
 * - conformsTo: [] An established standard to which the described resource conforms.
 *     Used to tag format specificities the data conforms to (ex. "sorted" for a GFF3).
 * - relation: [] A related resource.
 *     Recommended practice is to identify the related resource by means of a URI. If this is not possible or feasible, a string conforming to a formal identification system may be provided.
 *     Used to indicate derived files (ie. files generated using this file amongst others).
 * - source: [] A related resource from which the described resource is derived. URI.
 *     This property is intended to be used with non-literal values. The described resource may be derived from the related resource in whole or in part. Best practice is to identify the related resource by means of a URI or a string conforming to a formal identification system.
 *     Used for the original file name or to list files used to generate the current one.
 * - version: Version of current file or dataset.
 * - hasVersion: [] A related resource that is a version, edition, or adaptation of the described resource.
 *     Changes in version imply substantive changes in content rather than differences in format. This property is intended to be used with non-literal values. This property is an inverse property of Is Version Of.
 *     Used for other version URIs when a new version is released.
 * - isVersionOf: A related resource of which the described resource is a version, edition, or adaptation.
 *     Changes in version imply substantive changes in content rather than differences in format. This property is intended to be used with non-literal values. This property is an inverse property of Has Version.
 *     Used to link previous version when a new version is released.
 * - license: A legal document giving official permission to do something with the resource.
 *     Recommended practice is to identify the license document with a URI. If this is not possible or feasible, a literal value that identifies the license may be provided.
 * - rightsHolder: [] A person or organization owning or managing rights over the resource.
 *     Recommended practice is to refer to the rights holder with a URI. If this is not possible or feasible, a literal value that identifies the rights holder may be provided.
 *     Can be used for institutes.
 * - creator: [] An entity responsible for making the resource.
 *     Recommended practice is to identify the creator with a URI. If this is not possible or feasible, a literal value that identifies the creator may be provided.
 *     Used for authors.
 * - contributor: [] An entity responsible for making contributions to the resource.
 *     The guidelines for using names of persons or organizations as creators apply to contributors.
 *     Used for co-authors.
 * - description: An account of the resource.
 *     Description may include but is not limited to: an abstract, a table of contents, a graphical representation, or a free-text account of the resource.
 * - extent: The size or duration of the resource.
 *     Recommended practice is to specify the file size in megabytes and duration in ISO 8601 format.
 *     Used for the file size in Mb.
 * - isReferencedBy: [] A related resource that references, cites, or otherwise points to the described resource.
 *     This property is intended to be used with non-literal values. This property is an inverse property of References.
 *     When used in other datasets
 * - references: [] A related resource that is referenced, cited, or otherwise pointed to by the described resource.
 *     This property is intended to be used with non-literal values. This property is an inverse property of Is Referenced By.
 *     For dataset, to list used files or other resources.
 * - rights: Information about rights held in and over the resource.
 *     Typically, rights information includes a statement about various property rights associated with the resource, including intellectual property rights.
 * - accessRights: [] Information about who access the resource or an indication of its security status.
 *     Access Rights may include information regarding access or restrictions based on privacy, security, or other policies.
 *     Used to manage access restrictions.
 * - available: Date that the resource became or will become available.
 *     Recommended practice is to describe the date, date/time, or period of time as recommended for the property Date, of which this is a subproperty.
 * - tags: [] non-normalized terms with no values, attached to the data for searching and filtering.
 * - processor: The name of the main (or last) data processor (software, script, service) used to generate the file.
 * - processing: [] ordered list of treatments made to the source file to create this one.
 *     Each treatment is a command line used to generate the file with original paths replaced.
 * - isSource: to tell if the given file has not been altered by GenoRing.
 * - taxonomy: [] List of taxonomy identifiers used by the file or dataset.
 *     Numeric codes are public taxonomy identifier used on ENA and NCBI.
 *     5-capital letter codes are UniProt species identification codes.
 *     Others are considered as custom taxonomy names.
 *     See https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/docs/speclist.txt
 * - targetUri: to indicate the URI on which current file is about. Typically used for metadata files to point their corresponding data files.
 *
 * GenoRing uses data field. To bridges those fields with the Drupal system,
 * GenoRing fields use a generic definition array of the following structure:
 * - 'type': the field type machine name. Base types are defined in
 *   \Drupal\genoring\DataManager\DataManager::$dataFieldDefinitions:
 *   - 'boolean': for TRUE/FALSE values using a checkbox.
 *   - 'date': for dates.
 *   - 'daterange': for date ranges.
 *   - 'email': for e-mail addresses.
 *   - 'float': for floating point values.
 *   - 'genoring_input': for files or URIs.
 *   - 'integer': for integer values.
 *   - 'link': for URLs.
 *   - 'reference': for reference to other elements (entities).
 *   - 'string': for single-line string values.
 *   - 'text': for multi-line raw text values.
 *   - 'text_formatted': for multi-line HTML-formatted text values.
 *   - 'timestamp': for dates with time.
 *   - 'telephone': for telephone numbers.
 * - 'label': translatable and human readable label for the field.
 * - 'description': translatable field description.
 * - 'flags': changes the behavior of how the field is filled. See
 *   \Drupal\genoring\DataManager\DataManagerInterface::METADATA_FIELD_* for
 *   available flags.
 * - 'default': if METADATA_FIELD_AUTO flag is set, it should contain a
 *   default value or a callable function that would return the default
 *   value, given as parameter the field machine name and the data model
 *   machine name. If METADATA_FIELD_INHERIT flag is set, this value is
 *   expected to contain an ordered array of form field names this field
 *   could inherit from.
 * - 'overrides': an associative array of field overrides to apply to
 *   defaults provided by GenoRing in different keyed context:
 *   - 'form_element': overrides for a form field renderable array.
 *   - 'field_config': overrides for a form field configuration (as "Field"
 *     configuration type in configuration exports).
 *   - 'field_storage': overrides for a form field storage configuration (as
 *     "Field storage" configuration type in configuration exports).
 *   - 'view_display': overrides for a content field display (as what is
 *      defined for a field in the "content" part of an "Entity view
 *      display" configuration type in configuration exports).
 *   - 'form_display': overrides for a form field display (as what is
 *      defined for a field in the "content" part of an "Entity form
 *      display" configuration type in configuration exports).
 * - 'mapping': for mappable data fields, mapping definition arrays keyed by
 *   data model aggregator model identifiers:
 *   - 'mapping': source field or mapping expression.
 *   - 'data_processors': (optional) either a data processor ID or a data
 *     processor configuration array.
 *   - 'overrides': (optional) an external entity field mapper configuration
 *     replacing default mapping.
 *
 * @see https://www.dublincore.org/specifications/dublin-core/dcmi-terms/
 * @see https://edamontology.org/
 */
interface DataManagerInterface {

  /**
   * Flag to list subdirectories.
   *
   * @see self::listFiles
   */
  const LIST_SUBDIRS = 0b00000001;

  /**
   * Flag to also list hidden files (name starting by a dot).
   *
   * @see self::listFiles
   */
  const LIST_HIDDEN = 0b00000010;

  /**
   * Flag to also list symbolic links.
   *
   * @see self::listFiles
   */
  const LIST_LINKS = 0b00000100;

  /**
   * Flag to use case insensitive extension matching.
   *
   * @see self::listFiles
   */
  const LIST_EXT_CASE_INSENSITIVE = 0b00001000;

  /**
   * Flag to invert extension matching.
   *
   * @see self::listFiles
   */
  const LIST_EXT_INVERT_MATCH = 0b00010000;

  /**
   * Mode used to completely replace an existing metadata array.
   */
  const METADATA_REPLACE = 0;

  /**
   * Mode used to only add missing metadata values, existing values are left.
   */
  const METADATA_COMPLETE = 1;

  /**
   * Mode used to add missing metadata values and override existing ones.
   */
  const METADATA_OVERRIDE = 2;

  /**
   * Mode used to remove existing metadata.
   */
  const METADATA_REMOVE = 3;

  /**
   * Not editable field flag.
   *
   * When a field could be displayed but should not be edited by the user.
   */
  const METADATA_FIELD_NOT_EDITABLE = 0b00000001;

  /**
   * Hidden field flag.
   *
   * When a field should not be displayed but stored.
   */
  const METADATA_FIELD_HIDDEN = 0b00000010;

  /**
   * Inheritance field flag.
   *
   * When an empty field could get its value from fields with the same name
   * in a previous part of the global form.
   */
  const METADATA_FIELD_INHERIT = 0b00000100;

  /**
   * Auto-filled field flag.
   *
   * When a field will get its value automatically from a function if not set.
   */
  const METADATA_FIELD_AUTO = 0b00010000;

  /**
   * Empty-allowed field flag.
   *
   * When a field can be left empty.
   */
  const METADATA_FIELD_EMPTY = 0b00100000;

  /**
   * Creates the directory to place the given file if it does not exist.
   *
   * @param string $file_uri
   *   File or directory (ending by slash) URI.
   * @param bool $create_file
   *   Set to TRUE if an empty Drupal managed file should be created. If TRUE,
   *   no existing file would be replaced but a new one will be created instead
   *   if needed and $file_uri will be updated accordingly.
   *   If FALSE, no managed file will be created. A file with the same URI may
   *   already exist and will be ignored here.
   *
   * @return string
   *   The real path to the file.
   *
   * @throws \Drupal\genoring\Exception\DataManagerException
   *   Thrown if it failed to prepare link directory.
   */
  public function prepareFileDirectory(string &$file_uri, bool $create_file = FALSE) :string;

  /**
   * Generates a managed symlink.
   *
   * Generates a symlink to the given $target_uri file at the given $link_uri
   * place using a relative path. The generated symlink directory is created if
   * needed and the symlink becomes managed by Drupal system as a file
   * (returned instance).
   *
   * @param string $target_uri
   *   Target file URI.
   * @param string $link_uri
   *   Expected link URI.
   *
   * @return \Drupal\file\FileInterface|null
   *   The file entity corresponding to the symlink or NULL if the symlink was
   *   not created because the target and link path were the same.
   *
   * @throws \Drupal\genoring\Exception\DataManagerException
   *   Thrown if it failed to prepare link directory.
   */
  public function generateSymlink(string $target_uri, string $link_uri) :?FileInterface;

  /**
   * Returns the file matching the given metadata or NULL.
   *
   * @param array $metadata
   *   A metadata array.
   *
   * @return \Drupal\file\FileInterface|null
   *   A FileInterface instance if the file was found or NULL if the file was
   *   not found or if the metadata were not sufficient to locate the file.
   */
  public function getFile(array $metadata) :FileInterface|null;

  /**
   * Saves the given GenoRing input file or files as GenoRing managed files.
   *
   * @param array $input_value
   *   A genoring_input form element array of values.
   *
   * @return \Drupal\file\FileInterface[]
   *   An array of FileInterface instances if files were saved or an empty array
   *   if no file was saved.
   */
  public function saveGenoringInput(array $input_value) :array;

  /**
   * Returns the URI that should correspond to the metadata and type.
   *
   * @param array $metadata
   *   A metadata array. The metadata array must at least contain a key 'type'
   *   set to either 'file' or 'directory'.
   * @param bool $provide_default
   *   If TRUE and no file URI could be determined, it must return a default,
   *   non-empty URI.
   * @param array|null &$backtrace
   *   Gives a backtrace to track how data locators were used.
   *
   * @return string
   *   The URI. If the URI can not be determined and $provide_default is FALSE
   *   (default), an empty string is returned.
   *   If $provide_default is TRUE but it is still impossible to get a default
   *   base URI, and empty string is also returned.
   *
   * @throws \Drupal\genoring\Exception\MetadataException
   *   Thrown if the metadata does not meet requirements and $backtrace is
   *   provided.
   */
  public function locateUri(
    array $metadata,
    bool $provide_default = FALSE,
    array &$backtrace = [],
  ) :string;

  /**
   * Place an existing file on the file system according to the given metadata.
   *
   * @param \Drupal\file\FileInterface $file
   *   An existing file.
   * @param array $metadata
   *   A metadata array.
   *
   * @return \Drupal\file\FileInterface
   *   The new File object corresponding to the new file location.
   *
   * @throws \Drupal\Core\File\Exception\FileException
   *   Thrown when there is an error writing to the file system.
   * @throws \Drupal\Core\File\Exception\FileExistsException
   *   Thrown when the destination exists and $replace is set to
   *   FileSystemInterface::EXISTS_ERROR.
   * @throws \Drupal\Core\File\Exception\InvalidStreamWrapperException
   *   Thrown when the destination is an invalid stream wrapper.
   * @throws \Drupal\Core\Entity\EntityStorageException
   *   Thrown when there is an error saving the file.
   * @throws \Drupal\genoring\Exception\MetadataException
   *   Thrown if the metadata does not meet requirements.
   * @throws \Drupal\genoring\Exception\DataManagerException
   *   Thrown if the data locators failed to place the file.
   */
  public function placeFile(FileInterface $file, array $metadata) :FileInterface;

  /**
   * Place a set of files on the file system.
   *
   * @param array|string $file_list
   *   Either an array which keys are file URIs and values are metadata or a
   *   string containing a path to either a TSV or YAML file that contains a
   *   list of files with their metadata.
   *   The TSV file must have a first uncommented line with the column
   *   name corresponding to the metadata name and there must be at least a
   *   "uri" column and a "format" or "type" column.
   *   The YAML file should reflect the array structure: at first level, it must
   *   provide file URIs and as values an array of metadata for each file.
   *
   * @return \Drupal\file\FileInterface[]
   *   An array of File objects corresponding to the placed files.
   *
   * @throws \Drupal\Core\File\Exception\FileException
   *   Thrown when there is an error writing to the file system.
   * @throws \Drupal\Core\File\Exception\FileExistsException
   *   Thrown when the destination exists and $replace is set to
   *   FileSystemInterface::EXISTS_ERROR.
   * @throws \Drupal\Core\File\Exception\InvalidStreamWrapperException
   *   Thrown when the destination is an invalid stream wrapper.
   * @throws \Drupal\Core\Entity\EntityStorageException
   *   Thrown when there is an error saving the file.
   * @throws \Drupal\genoring\Exception\MetadataException
   *   Thrown if the metadata does not meet requirements.
   * @throws \Drupal\genoring\Exception\DataManagerException
   *   Thrown if the provided list is incorrect.
   */
  public function placeFiles(array|string $file_list) :array;

  /**
   * Place a set of files of an archive on the file system.
   *
   * @param string $zip_uri
   *   URI of the archive file. Supported archive types depend on the plugin.
   * @param array|string $metadata
   *   Metadata string or array.
   * @param bool $extract_files
   *   If TRUE files will be extracted.
   *
   * @return \Drupal\file\FileInterface[]
   *   An array of File objects corresponding to the placed files.
   *
   * @throws \Drupal\Core\File\Exception\FileException
   *   Thrown when there is an error writing to the file system.
   * @throws \Drupal\Core\File\Exception\FileExistsException
   *   Thrown when the destination exists and $replace is set to
   *   FileSystemInterface::EXISTS_ERROR.
   * @throws \Drupal\Core\File\Exception\InvalidStreamWrapperException
   *   Thrown when the destination is an invalid stream wrapper.
   * @throws \Drupal\Core\Entity\EntityStorageException
   *   Thrown when there is an error saving the file.
   * @throws \Drupal\genoring\Exception\MetadataException
   *   Thrown if the metadata does not meet requirements.
   * @throws \Drupal\genoring\Exception\DataManagerException
   *   Thrown if the provided archive is not valid.
   */
  public function placePackage(
    string $zip_uri,
    array|string $metadata,
    bool $extract_files = FALSE,
  ) :array;

  /**
   * List files in a given directory.
   *
   * @param string $directory
   *   Path or URI of directory to process.
   * @param string|array $extensions
   *   Filter on the given file extension list (array) or regular expression
   *   (string) *without* delimiters. Extensions should not include dots.
   *   If DataManagerInterface::LIST_EXT_INVERT_MATCH is used, only non-matching
   *   extensions will be listed, otherwise (default) only matching extensions
   *   will be listed.
   * @param int $flags
   *   Flags used to alter file listing behavior.
   *   - DataManagerInterface::LIST_SUBDIRS: also list subdirectories.
   *   - DataManagerInterface::LIST_HIDDEN: include hidden files.
   *   - DataManagerInterface::LIST_LINKS: include symbolic links (will happen a
   *     '*' to link file names).
   *   - DataManagerInterface::LIST_EXT_CASE_INSENSITIVE: ignore extension case.
   *   - DataManagerInterface::LIST_EXT_INVERT_MATCH: invert extension matching.
   *
   * @return array
   *   An array of form file names keyed by file path or URI (depending on
   *   $directory parameter).
   */
  public function listFiles(string $directory, string|array $extensions = [], int $flags = 0) :array;

  /**
   * Returns metadata associated to a file.
   *
   * @param string|FileInterface $file
   *   The file of interest, either its URI as a string or a FileInterface
   *   instance.
   *
   * @return array
   *   A metadata structure or an empty arry if no metadata is available.
   */
  public function getMetadata(string|FileInterface $file) :array;

  /**
   * Adjust and save new metadata associated to a file.
   *
   * @param string|FileInterface $file
   *   The file of interest, either its URI as a string or a FileInterface
   *   instance.
   * @param array $metadata
   *   New metadata array.
   * @param int $mode
   *   Mode to manage provided metadata. One of:
   *   - static::METADATA_REPLACE: replaces the whole metadata array if one
   *     (default).
   *   - static::METADATA_COMPLETE: complete by only adding metadata values that
   *     are missing and leaving existing ones as thery are.
   *   - static::METADATA_OVERRIDE: replaces existing values, add missing ones
   *     and leave other existing values that are not provided in the new array.
   *   - static::METADATA_REMOVE: remove existing metadata.
   *
   * @return self
   *   Current data manager instance.
   */
  public function setMetadata(
    string|FileInterface $file,
    array $metadata,
    int $mode = DataManagerInterface::METADATA_REPLACE,
  ) :DataManagerInterface;

  /**
   * Tells if the given metadata matches the given expression.
   *
   * This method can be used to check if a given metadata matching expression
   * matches the content of a metadata structure or not. Metadata matching
   * expressions have the following structure:
   * <EXPRESSION>[,<EXPRESSION>[,...]]
   * - <>: required.
   * - []: optional.
   * - |: alternatives.
   * - ...: repetition.
   * - EXPRESSION: <TERM|JSON_PATH>[=<VALUE|\[COMA_SEPARATED_VALUES\]>]
   * - TERM: a metadata term (key) name.
   * - JSON_PATH: a JSON path to locate the term in the metadata structure.
   * - VALUE: a value to match.
   * - COMA_SEPARATED_VALUES: a list of coma-separated values that must all be
   *     present in the array of values of the matched term.
   * Example of metadata structure:
   * @code
   * [
   *   'alpha' => 42,
   *   'beta' => [
   *     'gamma' => [501, 806],
   *     'omega' => [],
   *   ],
   *   'teta' => ['abc', 'de,f', 'ghi'],
   * ]
   * @endcode
   * Expression matching:
   * - 'alpha'
   * - 'alpha=42'
   * - 'alpha,beta'
   * - '$.alpha.gamma'
   * - '$.alpha.gamma=501'
   * - '$.alpha.gamma=[501]'
   * - '$.alpha.gamma=[501,806]'
   * - 'teta=[]'
   * - 'teta=[de\,f],$.beta.gamma,$.beta.omega'
   * Expression NOT matching:
   * - 'gamma' (no 'gamma' root element)
   * - 'alpha,psi' (no 'psi' root element)
   * - 'alpha=[]' ('alpha' is not an array of values)
   * - 'alpha=[42]' ('alpha' is not an array of values)
   * - '$.alpha.gamma=[501,42]' (no '42' element)
   * - '$.alpha.gamma=501,806' ('806' is considered as a second expression, no
   *   '806' root element)
   *
   * @param string $metadata_expression
   *   The metadata expression.
   * @param array $metadata
   *   The metadata to match.
   *
   * @return bool
   *   TRUE if there is a match, FALSE otherwise.
   */
  public function matchMetadata(
    string $metadata_expression,
    array $metadata,
  ) :bool;

  /**
   * Tells if any of the given expressions matches the given metadata.
   *
   * @param string|array $metadata_expressions
   *   An array of metadata expressions. If a string is provided, it will be
   *   exploded by lines. Each expression line can be prefixed by '-' for
   *   exclusion or '+' for inclusion (inclusion by default if not prefixed).
   * @param array $metadata
   *   The metadata to match.
   * @param array|null &$rules_backtrace
   *   Will contain rules backtrace if provided.
   *
   * @return bool
   *   TRUE if there is a match, FALSE otherwise.
   */
  public function matchMetadataAny(
    string|array $metadata_expressions,
    array $metadata,
    ?array &$rules_backtrace = NULL,
  ) :bool;

  /**
   * Returns the data model definitions.
   *
   * @return array
   *   Data model definitions keyed by model names.
   */
  public function getDataModelDefinitions() :array;

  /**
   * Returns the supported file types.
   *
   * @return array
   *   Supported file types are keys and values use the following structure:
   *  @code
   *    <file_type> => [
   *      'extensions' => [<list of extensions without dots>],
   *      'name' => <file type name>,
   *      'categories' => [[<main category>, <optional sub-category>]],
   *      'description' => <type description>,
   *      'mime' => <MIME type>,
   *      'plugins' => [<data processor plugins supporting this file type>],
   *      'specs' => <optional specification UTI>,
   *    ],
   *  @endcode
   */
  public function getFileTypes() :array;

  /**
   * Returns metadata fields info by data types.
   *
   * @param string $type
   *   The machine name of the type of data.
   *
   * @return array
   *   Metadata fields keyed by data types.
   */
  public function getMetadataFields(string $type) :array;

  /**
   * Returns available pre-processing operations.
   *
   * @param string $type
   *   The machine name of the type of data.
   *
   * @return array
   *   Available pre-processing operations keyed by data types and operation
   *   identifiers (sub-key). Each sub-array is an associative array of 2 keys:
   *   'form' that holds a sub-form rendering array, and 'sumbit' that holds the
   *   sub-form submit handler.
   *   See \Drupal\genoring\Event\DataTypeOperationsEvent for details.
   */
  public function getPreProcessingOperations(string $type) :array;

  /**
   * Generates a form field structure.
   *
   * @param string $field_name
   *   Field name/label.
   * @param array $field_def
   *   A GenoRing field definition structure as described in
   *   \Drupal\genoring\DataManager\DataManagerInterface documentation.
   *
   * @return array
   *   A form field structure.
   */
  public function generateFormField(
    string $field_name,
    array $field_def,
  ) :array;

  /**
   * Generates a field config structure.
   *
   * @param string $field_name
   *   Field name/label.
   * @param array $field_def
   *   A GenoRing field definition structure as described in
   *   \Drupal\genoring\DataManager\DataManagerInterface documentation.
   *
   * @return array
   *   A field config structure.
   */
  public function generateFieldConfig(
    string $field_name,
    array $field_def,
  ) :array;

  /**
   * Generates a field storage structure.
   *
   * @param string $field_name
   *   Field name/label.
   * @param array $field_def
   *   A GenoRing field definition structure as described in
   *   \Drupal\genoring\DataManager\DataManagerInterface documentation.
   *
   * @return array
   *   A field storage structure.
   */
  public function generateFieldStorage(
    string $field_name,
    array $field_def,
  ) :array;

  /**
   * Generates a field view display structure.
   *
   * @param string $field_name
   *   Field name/label.
   * @param array $field_def
   *   A GenoRing field definition structure as described in
   *   \Drupal\genoring\DataManager\DataManagerInterface documentation.
   *
   * @return array
   *   A field view display structure.
   */
  public function generateFieldViewDisplay(
    string $field_name,
    array $field_def,
  ) :array;

  /**
   * Generates a field form display structure.
   *
   * @param string $field_name
   *   Field name/label.
   * @param array $field_def
   *   A GenoRing field definition structure as described in
   *   \Drupal\genoring\DataManager\DataManagerInterface documentation.
   *
   * @return array
   *   A field form display structure.
   */
  public function generateFieldFormDisplay(
    string $field_name,
    array $field_def,
  ) :array;

  /**
   * Generates an external entity field mapping structure.
   *
   * @param string $field_name
   *   Field name to map used on source data or a mapping expression.
   * @param string|array $data_processors
   *   Array of data processor config to use.
   * @param array $overrides
   *   Mapping override array. Structure:
   *   @code
   *   [
   *     'id' => 'generic',
   *     'config' => [
   *       'property_mappings' => [
   *         //...
   *       ],
   *     ],
   *   ]
   *   @endcode
   *
   * @return array
   *   Returns a field mapping array suitable for data model field mappings.
   */
  public function generateFieldMapping(
    string $field_name,
    string|array $data_processors = [],
    array $overrides = [],
  ) :array;

}
