<?php

namespace Drupal\genoring\Element;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\Details;
use Drupal\Core\Render\Element\FormElementInterface;
use Drupal\file\Entity\File;
use Drupal\genoring\DataManager\DataManagerInterface;

/**
 * Provides a custom form element for 'genoring_input'.
 *
 * @FormElement("genoring_input")
 */
class GenoringInput extends Details implements FormElementInterface {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    $info = parent::getInfo();
    $info = [
      '#input' => TRUE,
      '#open' => TRUE,
      '#element_validate' => [[$class, 'validateGenoringInput']],
    ] + $info;
    $info['#process'][] = [$class, 'processGenoringInput'];
    // Unset '#value' otherwise ::valueCallback() is not called.
    unset($info['#value']);

    return $info;
  }

  /**
   * Process callback for 'genoring_input' element.
   *
   * @param array $element
   *   A reference to the form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   A reference to the complete form.
   */
  public static function processGenoringInput(
    array &$element,
    FormStateInterface $form_state,
    array &$complete_form,
  ) {
    $data_manager = \Drupal::service('genoring.data_manager');
    // Use or generate a unique HTML ID.
    $element_id = (
      $element['#attributes']['id'] ??= implode(
        '_',
        $element['#parents'] ?? [Html::getUniqueId('gi')]
      )
    );

    // Set up the form element for this widget.
    $element += [
      '#type' => 'details',
    ];
    $element['#tree'] = TRUE;
    // Single file?
    if (!empty($element['#fileset'])) {
      $element['import_method'] = [
        '#type' => 'select',
        '#title' => t('Data Import Method'),
        '#description' => t('Select the method to use to import data files.'),
        '#required' => $element['#required'] ?? FALSE,
        '#options' => [
          'single' => t('Provide a single data file'),
          'list' => t('Provide a list of data file URIs in a TSV or YAML file'),
          'archive' => t('Provide multiple data files in a zip or tar-gzip/bzip archive'),
        ],
        '#default_value' => 'single',
      ];
    }
    else {
      $element['import_method'] = [
        '#type' => 'hidden',
        '#value' => 'single',
      ];
    }
    $element['import_source'] = [
      '#type' => 'select',
      '#title' => t('Data Source'),
      '#description' => t('Select how the data file is provided.'),
      '#required' => $element['#required'],
      '#options' => [
        'browser' => t('Upload file from the web browser'),
        'updir' => t('Select file from the upload directory'),
        'uri' => t('Provide a file URI/URL'),
        'dataset' => t('Use a file from an existing dataset'),
      ],
      '#attributes' => [
        'data-genoring-selector' => 'isrc' . $element_id,
      ],
    ];

    // Get expected type.
    $extensions = [];
    $file_types = $data_manager->getFileTypes();
    if (!empty($element['#allowed_types'])) {
      if (!is_array($element['#allowed_types'])) {
        $element['#allowed_types'] = [$element['#allowed_types']];
      }
      foreach ($element['#allowed_types'] as $file_type) {
        if (!empty($file_types[$file_type])) {
          $extensions = array_merge($extensions, $file_types[$file_type]['extensions'] ?? []);
        }
      }
    }
    else {
      // Allow all extensions.
      $element['#allowed_types'] = [];
      foreach ($file_types as $file_type => $type_info) {
        $extensions = array_merge($extensions, $type_info['extensions'] ?? []);
      }
    }
    // We must wrap the managed_file field into a container to get arround
    // issue https://www.drupal.org/project/drupal/issues/2847425.
    $element['new_data_file_container'] = [
      '#type' => 'container',
      '#states' => [
        'visible' => [
          ':input[data-genoring-selector="isrc' . $element_id . '"]' => [
            'value' => 'browser',
          ],
        ],
      ],
      'new_data_file' => [
        '#type' => 'managed_file',
        '#title' => t('Upload new file'),
        '#upload_location' => 'private://upload',
        '#upload_validators' => [
          'FileExtension' => [
            'extensions' => implode(' ', array_unique($extensions)),
          ],
        ],
        '#default_value' => $form_state->getValue('new_data_file', ''),
      ],
    ];

    // Get the list of uploaded files.
    $local_file_options = $data_manager->listFiles(
      'private://upload/',
      $extensions,
      DataManagerInterface::LIST_SUBDIRS | DataManagerInterface::LIST_EXT_CASE_INSENSITIVE
    );
    $element['uploaded_data_file'] = [
      '#type' => 'select',
      '#title' => t('Select an already uploaded file'),
      '#options' => $local_file_options,
      '#empty_option' => t('- None -'),
      '#empty_value' => '',
      '#default_value' => $form_state->getValue('uploaded_data_file', ''),
      '#states' => [
        'visible' => [
          ':input[data-genoring-selector="isrc' . $element_id . '"]' => ['value' => 'updir'],
        ],
      ],
    ];
    $element['uri_data_file'] = [
      '#type' => 'textfield',
      '#title' => t('Enter a file URI/URL'),
      '#default_value' => $form_state->getValue('uri_data_file', ''),
      '#states' => [
        'visible' => [
          ':input[data-genoring-selector="isrc' . $element_id . '"]' => ['value' => 'uri'],
        ],
      ],
    ];

    $element['dataset_data_file'] = [
      '#type' => 'item',
      '#title' => t('File from an existing dataset: not implemented yet.'),
      '#states' => [
        'visible' => [
          ':input[data-genoring-selector="isrc' . $element_id . '"]' => ['value' => 'dataset'],
        ],
      ],
    ];
    $file_type = $form_state->getValue('file_type', '');
    $file_type_options = static::getFileTypeOptions($element, $form_state, $element['#allowed_types']);
    // Check if we have more than one option.
    if ((1 === count($file_type_options))
        && (!is_array(current($file_type_options))
            || (1 === count(current($file_type_options))))
    ) {
      $file_type = is_array(current($file_type_options))
        ? key(current($file_type_options))
        : key($file_type_options);
      $element['file_type'] = [
        '#type' => 'hidden',
        '#value' => $file_type,
      ];
    }
    else {
      // @todo Add ajax update to update metadata fields.
      $element['file_type'] = [
        '#type' => 'select',
        '#title' => t('File type'),
        '#description' => t('Select the file type.'),
        '#required' => $element['#required'],
        '#options' => static::getFileTypeOptions($element, $form_state, $element['#allowed_types']),
        '#default_value' => $file_type,
      ];
    }

    // Add metadata fields according to file type.
    // @todo Depending on 'import_source' value, this needs to be adjusted.
    $metadata_fields = $data_manager->getMetadataFields('file');
    if (!empty($file_type)) {
      $metadata_fields = NestedArray::mergeDeep(
        $metadata_fields,
        $data_manager->getMetadataFields($file_type)
      );
    }
    if (!empty($metadata_fields)) {
      $element['metadata'] = [
        '#type' => 'fieldset',
        '#title' => t('File metadata'),
      ];
      foreach ($metadata_fields as $field_name => $field_info) {
        $element['metadata'][$field_name] = $data_manager->generateFormField($field_name, $field_info);
      }
    }

    if (!empty($file_type)) {
      $operations = $data_manager->getPreProcessingOperations($file_type);
    }
    if (!empty($operations)) {
      // Processing.
      $element['processing'] = [
        '#type' => 'fieldset',
        '#title' => t('Data Processing'),
      ];
      foreach ($operations as $operation_id => $operation) {
        $element['processing'][$operation_id] = $operation['form'];
      }
    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    // We return an empty array instead of NULL to avoid keeping the given input
    // structure that has non-hash prefixed keys with values that will be
    // treaded as invalid renderable arrays and generate errors.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public static function validateGenoringInput(&$element, FormStateInterface $form_state, &$complete_form) {
    // Get services and config.
    $data_manager = \Drupal::service('genoring.data_manager');
    $file_repository = \Drupal::service('file.repository');
    $file_system = \Drupal::service('file_system');
    $configuration = $data_manager->getConfiguration();

    // Initializes current values.
    $values = $form_state->getValue($element['#parents'], []);
    $values['uris'] ??= [];
    $values['metadata'] ??= [];

    // Try to get file identifier "fid".
    $source = $element['import_source']['#value'] ?? NULL;
    if ('browser' == $source) {
      $fid = $element['new_data_file_container']['new_data_file']['#value']['fids'][0] ?? NULL;
      if (!empty($fid)) {
        // We got a local managed file, update URI value.
        $file = File::load($fid);
        $values['uris'] = [$file->getFileUri()];
      }
    }
    elseif ('updir' == $source) {
      $uploaded_file = $element['uploaded_data_file']['#value'] ?? NULL;
      // Make sure it is in GenoRing upload directory.
      $uploaded_file_realpath = $file_system->realpath($uploaded_file);
      $upload_realpath = $file_system->realpath($configuration['upload_uri'] ?? '#' ?: '#');
      if (!empty($uploaded_file)) {
        if ($upload_realpath
          && $uploaded_file_realpath
          && (0 === strpos($uploaded_file_realpath, $upload_realpath))
        ) {
          $values['uris'] = [$uploaded_file];
        }
        else {
          $form_state->setError($element, 'Invalid uploaded file.');
        }
      }
    }
    elseif ('uri' == $source) {
      if (!empty($values['uri_data_file'])) {
        if (preg_match('~^(?:https?|public|private)://~', $values['uri_data_file'])) {
          $values['uris'] = [$values['uri_data_file']];
        }
        else {
          $form_state->setError($element, 'Invalid file URI.');
        }
      }
    }
    elseif ('dataset' == $source) {
      // @todo Get source file fid.
    }

    // Set new values.
    $form_state->setValueForElement($element, $values);

    // Process file if needed.
    if (!empty($values['processing']) && !empty($values['file_type'])) {
      $operations = $data_manager->getPreProcessingOperations($values['file_type']);
    }
    if (!empty($operations)) {
      foreach ($operations as $operation_id => $operation) {
        // @todo Check if we got a subform for the operation.
        // Create sub-from and sub-form state.
        // Set 'uri' state.
        // Call submit handler $operation['submit'].
        // Check if URI changed and update if needed.
        // Save processed file metadata.
      }
    }
  }

  /**
   * List supported file types.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $type_filter
   *   List of allowed types.
   *
   * @return array
   *   An array of file type options for a 'select' form element.
   */
  protected static function getFileTypeOptions(array &$element, FormStateInterface $form_state, array $type_filter = []) :array {
    $type_filter = array_fill_keys(array_values($type_filter), 1);
    $data_manager = \Drupal::service('genoring.data_manager');
    $file_types = $data_manager->getFileTypes();
    $file_type_options = [];
    foreach ($file_types as $file_type => $file_type_definition) {
      if (!empty($type_filter) && empty($type_filter[$file_type])) {
        // Skip file types not allowed.
        continue;
      }
      $categories = $file_type_definition['categories'];
      if (!empty($categories)) {
        foreach ($categories as $category) {
          // Translate categories here and cast them to string because the
          // "select" form element needs options to be strings in order to work
          // properly, otherwise, objects are treated differently.
          $category = implode('/', array_map(
            function ($cat) {
              // The cat(egory) string comes from a literal string.
              // phpcs:ignore
              return (string) t($cat);
            },
            $category
          ));
          NestedArray::setValue(
            $file_type_options,
            [$category, $file_type],
            // The type definition name comes from a literal string.
            // phpcs:ignore
            (string) t($file_type_definition['name'])
          );
        }
      }
      else {
        $file_type_options[$file_type] = $file_type_definition['name'];
      }
    }

    return $file_type_options;
  }

}
