<?php

namespace Drupal\genoring\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides a 'genoring_license' form element.
 *
 * @FormElement("genoring_license")
 */
class License extends FormElement {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processLicense'],
      ],
      '#element_validate' => [[$class, 'validateLicense']],
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * Process callback for the license element.
   */
  public static function processLicense(array &$element, FormStateInterface $form_state, array &$complete_form) {
    // Use or generate a unique HTML ID.
    $element_id = (
      $element['#attributes']['id'] ??= implode(
        '_',
        $element['#parents'] ?? [Html::getUniqueId('gi')]
      )
    );

    // Set up the form element for this widget.
    $element += [
      '#type' => 'container',
    ];
    $licenses = static::getLicenses();
    $default_license = $element['#default_value'] ?? NULL;
    $other_license = '';
    if (isset($default_license)) {
      if (!array_key_exists($default_license, $default_license)) {
        $other_license = $default_license;
        $default_license = 'Other';
      }
    }
    else {
      $default_license = 'Unknown';
    }

    $element['common'] = [
      '#type' => 'select',
      '#options' => $licenses + [
        'Other' => t('Other'),
      ],
      '#default_value' => $default_license,
      '#attributes' => [
        'data-genoring-selector' => 'license' . $element_id,
      ],
    ];

    $element['other'] = [
      '#type' => 'textfield',
      '#title' => t('Enter license type'),
      '#default_value' => $other_license,
      '#states' => [
        'visible' => [
          ':input[data-genoring-selector="license' . $element_id . '"]' => [
            'value' => 'Other',
          ],
        ],
      ],
    ];

    return $element;
  }

  /**
   * Returns an array of selectable licenses.
   */
  public static function getLicenses() :array {
    $licenses = [
      'Unknown' => t('Unknown'),
      'CC0' => t('CC0 (No Rights Reserved, public domain)'),
      'CC-by' => t('CC-by (Attribution)'),
      'CC-by-sa' => t('CC-by-sa (Attribution-ShareAlike)'),
      'CC-by-nc' => t('CC-by-nc (Attribution-NonCommercial)'),
      'CC-by-nc-sa' => t('CC-by-nc-sa (Attribution-NonCommercial-ShareAlike)'),
      'CC-by-nd' => t('CC-by-nd (Attribution-NoDerivatives)'),
      'CC-by-nc-nd' => t('CC-by-nc-nd (Attribution-NonCommercial-NoDerivatives)'),
      'GNU AGPLv3' => t('GNU Affero General Public License v3.0'),
      'GNU GPLv3' => t('GNU General Public License v3.0'),
      'GNU LGPLv3' => t('GNU Lesser General Public License v3.0'),
      'Mozilla Public License 2.0' => t('Mozilla Public License 2.0'),
      'Apache License 2.0' => t('Apache License 2.0'),
      'MIT License' => t('MIT License'),
      'The Unlicense' => t('The Unlicense'),
    ];
    // @todo Add event to alter licence list.
    return $licenses;
  }

  /**
   * {@inheritdoc}
   */
  public static function validateLicense(&$element, FormStateInterface $form_state, &$complete_form) {
    // Initializes current values.
    $values = $form_state->getValue($element['#parents'], '');

    if (isset($values['common']) && isset($values['other'])) {
      if ($values['common'] !== 'Other') {
        $values = $values['common'];
      }
      else {
        $values = $values['other'];
      }
    }

    // Set new values.
    $form_state->setValueForElement($element, $values);
  }

}
