<?php

namespace Drupal\genoring\Event;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Component\Utility\NestedArray;

/**
 * The Data Models Event.
 *
 * The DataModelsEvent gather information from event handlers on available data
 * models.
 *
 * Each event handler can add new models or complete existing ones.
 */
class DataModelsEvent extends Event {

  /**
   * The data model definitions array.
   *
   * The data model definitions array holds data model defninitions keyed by
   * their machine names. Each data model definition array provides information
   * on dataset fields (ie. input files and metadata) and data model fields with
   * their mapping on dataset fields. Data model arrays use the following
   * structure:
   * - 'label': the translatable label of the data model.
   * - 'description': the translatable description of the data model.
   * - 'dataset_fields': a set of dataset GenoRing field definitions keyed by
   *   their machine names. These are mostly used to collect dataset metadata
   *   and base files.
   * - 'model_fields': a set of model GenoRing field definitions keyed by their
   *   machine names. Those field definitions include mapping definitions and
   *   are mostly used as displayed fields of the data model.
   * - 'sources': a set of external entity data model aggregator sources
   *   definitions keyed by their machine name.
   * - 'form': form handlers (callable):
   *   - 'alter': called to alter the form with the $form and $form_state
   *     parameters. Form should be altered using reference. No return value is
   *     expected.
   *   - 'validate': called to validate the form with the $form and $form_state
   *     parameters.
   *   - 'submit': called when the form is submitted, after the dataset and
   *     corresponding external entity instance have been created. The external
   *     entity instance may be incomplete and this handler can be used to
   *     complete what is missing. Parameters:
   *     - array &$form: the form structure.
   *     - \Drupal\Core\Form\FormStateInterface $form_state: the form state.
   *     - string $model_id: current data model identifier.
   *     - \Drupal\genoring\Entity\DatasetInterface $dataset: dataset instance.
   *     - \Drupal\external_entities\Entity\ExternalEntityTypeInterface $xntt:
   *       corresponding data model external entity type instance.
   *
   * @var array
   *   A data model definition array as described above.
   */
  protected $dataModelDefinitions;

  /**
   * Constructs a data model event object.
   */
  public function __construct() {
    $this->dataModelDefinitions = [];
  }

  /**
   * Gets data models.
   *
   * @return array
   *   The data models.
   */
  public function getDataModelDefinitions() :array {
    return $this->dataModelDefinitions;
  }

  /**
   * Add data models.
   *
   * @param array $data_models
   *   The data models to add or complete.
   * @param string $merge_mode
   *   Merge mode. Can be either 'replace' (default) to replace existing
   *   definitions or 'keep' to keep existing definitions and just append
   *   missing elements.
   *
   * @return self
   *   Current instance.
   */
  public function mergeDataModelDefinitions(
    array $data_models,
    string $merge_mode = 'replace',
  ) :self {
    if ('replace' == $merge_mode) {
      // Replace existing definitions.
      $this->dataModelDefinitions = NestedArray::mergeDeep(
        $this->dataModelDefinitions,
        $data_models
      );
    }
    else {
      // "keep" mode.
      $this->dataModelDefinitions = NestedArray::mergeDeep(
        $data_models,
        $this->dataModelDefinitions
      );
    }
    return $this;
  }

}
