<?php

namespace Drupal\genoring\Event;

use Drupal\Component\EventDispatcher\Event;

/**
 * The type pre-processing operations Event.
 *
 * Provides a way to gather pre-processing operations related to a given file
 * type.
 */
class DataTypeOperationsEvent extends Event {

  /**
   * The available pre-processing operations keyed by data types.
   *
   * @var array
   */
  protected $operations;

  /**
   * Constructs a compatible data model event object.
   */
  public function __construct() {
    $this->operations = [];
  }

  /**
   * Gets the list of available operations.
   *
   * @return array
   *   A list of operations keyed by data types. Each operation array has 2
   *   keys: 'form' with the operation form elements and 'submit' with the
   *   corresponding submit handler.
   */
  public function getPreProcessingOperations() :array {
    return $this->operations;
  }

  /**
   * Add a pre-processing operation.
   *
   * @param string $type
   *   The machine name of the type of data.
   * @param string $operation_id
   *   An identifier for the type of operation.
   * @param array $operation_form
   *   The form render array structure to use.
   * @param array $operation_submit
   *   The submit handler that will be called to manage the operation. It must
   *   use the same structure as the one provided to #submit for buttons (ie.
   *   ['class', 'methodName']). It will be called with a sub-form and a
   *   sub-form state with the state 'uri' set to the URI of the file to
   *   process and it should change that state to the new processed file URI, if
   *   a processing has been done. The state 'metadata' will also be set for the
   *   given file, it could also be used and should be updated accordingly.
   *
   * @return self
   *   Current instance.
   */
  public function addPreProcessingOperation(
    string $type,
    string $operation_id,
    array $operation_form,
    array $operation_submit,
  ) :self {
    $this->operations[$type] ??= [];
    $this->operations[$type][$operation_id] = [
      'form' => $operation_form,
      'submit' => $operation_submit,
    ];
    return $this;
  }

}
