<?php

namespace Drupal\genoring\EventSubscriber;

use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\external_entities\Entity\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\ExternalEntities\DataAggregator\GroupAggregator;
use Drupal\genoring\Entity\DatasetInterface;
use Drupal\genoring\Event\DataModelManagementEvent;
use Drupal\genoring\Event\DataModelsEvent;
use Drupal\genoring\Event\DataTypeMetadataEvent;
use Drupal\genoring\Event\FileTypesEvent;
use Drupal\genoring\Event\GenoringEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * GenoRing event subscriber to handle default behaviors.
 */
class GenoringEventsSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * Constructs a GenoringEventsSubscriber object.
   *
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   */
  public function __construct(TranslationInterface $string_translation) {
    $this->stringTranslation = $string_translation;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[GenoringEvents::FILE_TYPES][] = 'genoringFileTypes';
    $events[GenoringEvents::DATA_TYPE_METADATA][] = 'genoringDataTypeMetadata';
    $events[GenoringEvents::DATA_MODELS][] = 'genoringDataModels';
    $events[GenoringEvents::DATA_MODEL_MANAGEMENT][] = 'genoringDataModelManagement';
    return $events;
  }

  /**
   * Add default data models.
   *
   * Check available files and provide default form elements for default data
   * models.
   *
   * @param \Drupal\genoring\Event\DataModelsEvent $event
   *   The event.
   */
  public function genoringDataModels(DataModelsEvent $event) {
    $definitions['assembly'] = [
      'label' => $this->t('Genome Assembly'),
      'description' => $this->t('A genome assembly.'),
      'dataset_fields' => [
        'version' => [
          'type' => 'string',
          'label' => $this->t('Version'),
        ],
        'license' => [
          'type' => 'license',
          'label' => $this->t('License'),
        ],
        'rightsHolder' => [
          'type' => 'string',
          'label' => $this->t('Institute/rights owner'),
        ],
        'description' => [
          'type' => 'text',
          'label' => $this->t('Description'),
        ],
        'contributor' => [
          'type' => 'text',
          'label' => $this->t('Contributors'),
        ],
        'taxonomy' => [
          'type' => 'taxonomy',
          'label' => $this->t('Data-related taxonomy'),
        ],
        'genome_fasta' => [
          'type' => 'genoring_input',
          'label' => $this->t('FASTA file'),
          'overrides' => [
            'form_element' => [
              '#allowed_types' => 'fasta',
              '#tags' => [
                'required' => ['dna', 'sequences'],
              ],
            ],
          ],
        ],
        'genome_gff' => [
          'type' => 'genoring_input',
          'label' => $this->t('GFF3 file'),
          'overrides' => [
            'form_element' => [
              '#allowed_types' => 'gff3',
              '#tags' => [
                'required' => [],
              ],
            ],
          ],
        ],
      ],
      'model_fields' => [
        'id' => [
          'mapping' => [
            'assembly' => [
              'mapping' => 'id',
            ],
          ],
        ],
        'title' => [
          'mapping' => [
            'assembly' => [
              'mapping' => 'name',
            ],
          ],
        ],
        'genoring_genome_version' => [
          'type' => 'string',
          'label' => $this->t('Assembly version'),
          'mapping' => [
            'assembly' => [
              'mapping' => 'version',
              // 'data_processors' => [],
              // 'overrides' => [],
            ],
          ],
        ],
        'genoring_genome_fasta' => [
          'type' => 'string',
          'label' => $this->t('Genome FASTA'),
          'mapping' => [
            'assembly' => [
              'mapping' => 'fasta',
            ],
          ],
        ],
        'genoring_genome_gff3' => [
          'type' => 'string',
          'label' => $this->t('Genome GFF3'),
          'mapping' => [
            'assembly' => [
              'mapping' => 'gff3',
            ],
          ],
        ],
      ],
      'sources' => [
        'assembly' => [
          [
            'id' => 'xntttsv',
            'config' => [
              'root' => '',
              'headers' => 'comment',
              'normalize_headers' => 0,
              'comment_prefix' => '#',
              'separator' => "\t",
              'quoted' => 1,
              'generate_id' => 0,
            ],
            'aggr' => [
              // Users are not allowed to edit this TSV.
              'mode' => GroupAggregator::STORAGE_CLIENT_MODE_READONLY,
            ],
          ],
        ],
      ],
      'form' => [
        'submit' => [static::class, 'submitAssemblyForm'],
      ],
    ];
    // Disabled for alpha.
    // @code
    // $definitions['gene'] = [
    //   'label' => $this->t('Gene set'),
    //   'description' => $this->t(
    //     'One or more genes corresponding to a gene data model.'
    //   ),
    //   'dataset_fields' => [
    //     'version' => [
    //       'type' => 'string',
    //       'label' => $this->t('Version'),
    //     ],
    //     'license' => [
    //       'type' => 'license',
    //       'label' => $this->t('License'),
    //     ],
    //     'rightsHolder' => [
    //       'type' => 'string',
    //       'label' => $this->t('Institute/rights owner'),
    //     ],
    //     'description' => [
    //       'type' => 'text',
    //       'label' => $this->t('Description'),
    //     ],
    //     'contributor' => [
    //       'type' => 'text',
    //       'label' => $this->t('Contributors'),
    //     ],
    //     'taxonomy' => [
    //       'type' => 'taxonomy',
    //       'label' => $this->t('Data-related taxonomy'),
    //     ],
    //     'gene_fasta' => [
    //       'type' => 'genoring_input',
    //       'label' => $this->t('Gene FASTA file'),
    //       'overrides' => [
    //         'form_element' => [
    //           '#allowed_types' => 'fasta',
    //           '#tags' => [
    //             'required' => ['dna', 'sequences'],
    //           ],
    //         ],
    //       ],
    //     ],
    //     'protein_fasta' => [
    //       'type' => 'genoring_input',
    //       'label' => $this->t('Protein FASTA file'),
    //       'overrides' => [
    //         'form_element' => [
    //           '#allowed_types' => 'fasta',
    //           '#tags' => [
    //             'required' => ['protein', 'sequences'],
    //           ],
    //         ],
    //       ],
    //     ],
    //     'genome_gff' => [
    //       'type' => 'genoring_input',
    //       'label' => $this->t('GFF3 file'),
    //       'overrides' => [
    //         'form_element' => [
    //           '#allowed_types' => 'gff3',
    //           '#tags' => [
    //             'required' => [],
    //           ],
    //         ],
    //       ],
    //     ],
    //   ],
    // ];
    // @endcode

    $event->mergeDataModelDefinitions($definitions);
  }

  /**
   * Genome form validation.
   */
  public static function submitAssemblyForm(array &$form, FormStateInterface $form_state, string $model_id, DatasetInterface $dataset, ExternalEntityTypeInterface $xntt) {
    $data_manager = \Drupal::service('genoring.data_manager');
    $model_definition = $form_state->get('models.' . $model_id);

    // Get FASTA file URI.
    $genome_fasta_input = $form_state->getValue('genome_fasta', []);
    $fasta_file = current($data_manager->saveGenoringInput($genome_fasta_input));
    if (!empty($fasta_file)) {
      // Place file to the dataset directory.
      $metadata = $genome_fasta_input['metadata'];
      $metadata['dataset'] = $dataset->id();
      $metadata['isPartOf'] = 'dataset-' . $dataset->id();
      $fasta_file = $data_manager->placeFile($fasta_file, $metadata);
      $dataset->addDataFile($fasta_file);
    }

    // Get GFF3 file URI.
    $genome_gff_input = $form_state->getValue('genome_gff', []);
    $gff_file = current($data_manager->saveGenoringInput($genome_gff_input));
    if (!empty($gff_file)) {
      // Place file to the dataset directory.
      $metadata = $genome_gff_input['metadata'];
      $metadata['dataset'] = $dataset->id();
      $metadata['isPartOf'] = 'dataset-' . $dataset->id();
      $gff_file = $data_manager->placeFile($gff_file, $metadata);
      $dataset->addDataFile($gff_file);
    }

    // Generate assembly TSV file.
    $metadata = [
      'type' => 'file',
      'filename' => $dataset->id() . '_assembly.tsv',
      'dataset' => $dataset->id(),
      'isPartOf' => 'dataset-' . $dataset->id(),
    ];
    foreach ($model_definition['dataset_fields'] as $field_name => $field_info) {
      if ('genoring_input' != $field_info['type']) {
        $metadata[$field_name] = $form_state->getValue($field_name);
      }
    }
    $assembly_file_uri = $data_manager->locateUri($metadata);
    $data =
      "#id\tname\tversion\tfasta\tgff3\n"
      . $dataset->id()
      . "\t"
      . $dataset->label()
      . "\t"
      . $metadata['version']
      . "\t"
      . (!empty($fasta_file) ? $fasta_file->getFileUri() : '')
      . "\t"
      . (!empty($gff_file) ? $gff_file->getFileUri() : '')
      . "\n";
    $file_repository = \Drupal::service('file.repository');
    $file_system = \Drupal::service('file_system');
    $dataset_directory = dirname($assembly_file_uri);
    if (!$file_system->prepareDirectory($dataset_directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger->error('Failed to create directory @directory.', ['@directory' => $dataset_directory]);
      throw new DataManagerException('Failed to create directory ' . $dataset_directory . '.');
    }
    $file = $file_repository->writeData($data, $assembly_file_uri, FileExists::Replace);
    $file->setPermanent();
    $file->save();

    // Add TSV file to dataset.
    $dataset->addDataFileUri($assembly_file_uri);

    // Update xntt assembly storage client mapping.
    $aggr_config = $xntt->getDataAggregatorConfig();
    $aggr_config['storage_clients'][0]['config']['root'] = $dataset_directory;
    $aggr_config['storage_clients'][0]['config']['structure'] = basename($assembly_file_uri);
    $xntt->setDataAggregatorConfig($aggr_config);
    $xntt->save();
  }

  /**
   * Provides default action to available data models.
   *
   * @param \Drupal\genoring\Event\DataModelManagementEvent $event
   *   The event.
   */
  public function genoringDataModelManagement(DataModelManagementEvent $event) {
    $data_model = $event->getDataModel();
    $management_form = $event->getManagementForm();
    // @todo Implement UI to edit metadata and change data files.
    $event->setManagementForm($management_form);
  }

  /**
   * Provides default file types.
   *
   * Supported:
   * - Sequences/alignments:
   *   - FASTA files in various flavors
   * - Media:
   *   - PNG images
   *   - JPEG images
   * - Others:
   *   - GFF3 files
   *   - TSV files.
   *
   * // Not added yet: CSV, FASTQ, SAM, BAM, CRAM, stockholm (.sto, .stk), VCF,
   * // BCF, GTF, GFF2, Newick (.nwk, .nhx, .tree), NEXUS, PHYLIP (.phy, .ph),
   * // Clustal, PhyloXML, XML, YAML, TXT, MD, RDF, DOC(x), XLS(x), LOG, PPT,
   * // PDF, EMBL, EMBOSS, BLAST outputs, BED, SVG, TIFF, GIF, JSON, ZIP, TGZ,
   * // GZ, BZ, TBZ, TAR, LaTeX (.tex), SQL, DIFF, ...
   *
   * @param \Drupal\genoring\Event\FileTypesEvent $event
   *   The event.
   */
  public function genoringFileTypes(FileTypesEvent $event) {
    // @todo To complete.
    $event->addFileTypes(
      [
        // Sequences and alignments.
        'fasta' => [
          'extensions' => ['fa', 'fas', 'fasta', 'ffn', 'fna', 'frn', 'faa', 'seq'],
          'name' => 'FASTA sequences',
          'categories' => [['Sequences']],
          'description' => 'One or more nucleic or amino acid sequences in FASTA format.',
          'mime' => 'text/plain',
          'plugins' => [],
          'specs' => 'https://en.wikipedia.org/wiki/FASTA_format',
        ],
        'fasta_alignment' => [
          'extensions' => ['fa', 'fas', 'fasta', 'ffn', 'fna', 'frn', 'faa', 'seq', 'fsa'],
          'name' => 'FASTA alignment',
          'categories' => [['Alignments']],
          'description' => 'Multiple nucleic or amino acid sequence alignment in FASTA format.',
          'mime' => 'text/plain',
          'plugins' => [],
          'specs' => 'https://en.wikipedia.org/wiki/FASTA_format',
        ],
        // Media.
        'jpeg' => [
          'extensions' => ['jpg', 'jpeg'],
          'name' => 'JPEG image',
          'categories' => [['Media', 'Images']],
          'description' => 'Joint Photographic Experts Group image.',
          'mime' => 'image/jpeg',
          'plugins' => [],
        ],
        'png' => [
          'extensions' => ['png'],
          'name' => 'PNG image',
          'categories' => [['Media', 'Images']],
          'description' => 'Portable Network Graphics image.',
          'mime' => 'image/png',
          'plugins' => [],
        ],
        'text' => [
          'extensions' => ['txt', 'md'],
          'name' => 'Text file',
          'categories' => [['Media', 'Documents']],
          'description' => 'Basic text file or markdown.',
          'mime' => 'text/plain',
          'plugins' => [],
        ],
        // Others.
        'gff3' => [
          'extensions' => ['gff', 'gff3'],
          'name' => 'GFF3',
          'categories' => [['Others']],
          'description' => 'Generic Feature Format Version 3 (GFF3).',
          'mime' => 'text/plain',
          'plugins' => [],
          'specs' => 'https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md',
        ],
        'tsv' => [
          'extensions' => ['tsv', 'tab'],
          'name' => 'TSV',
          'categories' => [['Others']],
          'description' => 'Tabulation-separated value file.',
          'mime' => 'text/tab-separated-values',
          'plugins' => [],
        ],
        'tsv_germplasm' => [
          'extensions' => ['tsv', 'tab'],
          'name' => 'Germplasm TSV',
          'categories' => [['Others']],
          'description' => 'Tabulation-separated value file containing germplasm data.',
          'mime' => 'text/tab-separated-values',
          'plugins' => [],
        ],
      ]
    );
  }

  /**
   * Event handler for data type metadata event.
   *
   *  Generates metadata fields according to the data type.
   *
   * @param \Drupal\genoring\Event\DataTypeMetadataEvent $event
   *   The event.
   */
  public function genoringDataTypeMetadata(DataTypeMetadataEvent $event) {
    $event->addMetadataField(
      'file',
      'created',
      [
        'type' => 'datetime',
        'label' => $this->t('File creation date'),
      ],
    );
    $event->addMetadataField(
      'file',
      'version',
      [
        'type' => 'string',
        'label' => $this->t('File version'),
      ],
    );
    $event->addMetadataField(
      'file',
      'license',
      [
        'type' => 'license',
        'label' => $this->t('License'),
      ],
    );
    $event->addMetadataField(
      'file',
      'rightsHolder',
      [
        'type' => 'string',
        'label' => $this->t('Institute/rights owner'),
      ],
    );
    $event->addMetadataField(
      'file',
      'creator',
      [
        'type' => 'string',
        'label' => $this->t('Data file creator'),
      ],
    );
    $event->addMetadataField(
      'file',
      'description',
      [
        'type' => 'text',
        'label' => $this->t('Description'),
      ],
    );
    $event->addMetadataField(
      'file',
      'contributor',
      [
        'type' => 'text',
        'label' => $this->t('Contributors'),
      ],
    );
    $event->addMetadataField(
      'file',
      'taxonomy',
      [
        'type' => 'taxonomy',
        'label' => $this->t('Data-related taxonomy'),
      ],
    );
  }

}
