<?php

namespace Drupal\genoring\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\genoring\DataManager\DataManager;
use Drupal\genoring\DataManager\DataManagerInterface;
use Drupal\genoring\Exception\MetadataException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Exception\ParseException;
use Symfony\Component\Yaml\Yaml;

/**
 * Provides the data manager configuration form.
 */
class DataLocatorForm extends FormBase {

  /**
   * The data manager.
   *
   * @var \Drupal\genoring\DataManager\DataManagerInterface
   */
  protected $dataManager;

  /**
   * Constructs an DataLocatorForm object.
   *
   * @param \Drupal\genoring\DataManager\DataManagerInterface $data_manager
   *   The data processor manager.
   */
  public function __construct(
    DataManagerInterface $data_manager,
  ) {
    $this->dataManager = $data_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('genoring.data_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'genoring_data_locator_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Disable client-side validation.
    $form['#attributes']['novalidate'] = 'novalidate';

    // Set a predictible id.
    $form['#attributes'] += [
      'id' => 'genoring_dl',
    ];

    $form['info'] = [
      '#type' => 'item',
      '#markup' => $this->t(
        "The data locator is used to place files and directories in the file system accoring to their metadata. On this interface, you can configure and test the correct behaviour of your data locator settings."
      ),
    ];

    $form = $this->dataManager->buildDataLocatorConfigurationForm($form, $form_state);

    $test_result = $form_state->get('result');
    $form['data_locators']['#open'] = !isset($test_result);
    $form['tests'] = [
      '#type' => 'details',
      '#title' => $this->t('Test settings'),
      '#open' => isset($test_result),
      '#tree' => FALSE,
    ];

    $metadata_example = "type: file\nfilename: pahang2.fasta\nformat: metadata\ntargetUri: private://genoring/gen/MUSAC/pahang_v2_0/pahang2.fasta\ntargetType: file\nuri: private://upload/pahang2.fasta\nisSource: false\ntaxonomy: Musa acuminata\nassembly: Pahang v2.0";
    $form['tests']['metadata'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Input metadata'),
      '#description' => $this->t('Provide metadata in YAML format.'),
      '#required' => TRUE,
      '#attributes' => [
        'placeholder' => "Example:\n$metadata_example",
      ],
    ];
    $form['tests']['provide_default'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Compute a default URI if metadata does not match any rule'),
    ];

    $backtrace = $form_state->get('backtrace');
    if (!empty($test_result) || !empty($backtrace)) {
      $form['tests']['result'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Result'),
      ];
      if (!empty($test_result)) {
        $form['tests']['result']['uri'] = [
          '#type' => 'item',
          '#title' => $this->t('URI:'),
          '#markup' => "<strong>$test_result</strong>",
          '#wrapper_attributes' => ['class' => ['container-inline']],
        ];
      }
      else {
        $form['tests']['result']['uri'] = [
          '#type' => 'item',
          '#title' => $this->t('URI:'),
          '#markup' => "No match!",
          '#wrapper_attributes' => ['class' => ['container-inline']],
        ];
      }
      if (!empty($backtrace)) {
        $data_manager_config = \Drupal::config(DataManager::CONFIG_KEY)->get() ?? [];
        $locator_descriptions = array_values(
          array_map(
            function ($locator) {
              return $locator['id']
                . ': '
                . preg_replace('/\n.*/', '', $locator['notes'] ?? '');
            },
            $data_manager_config['data_locators'] ?? []
          )
        );
        $backtrace_text = "<ul>\n";
        foreach ($backtrace as $locator_index => $locator_trace) {
          $backtrace_text .= "  <li>\n    $locator_index (" . $locator_descriptions[$locator_index] . "): " . ($locator_trace['uri'] ?? '-') . "\n";
          if (!empty($locator_trace['rules'])) {
            $backtrace_text .= "    <ul>\n";
            foreach ($locator_trace['rules'] as $rule => $rule_result) {
              $backtrace_text .= "      <li>$rule: $rule_result</li>\n";
            }
            $backtrace_text .= "    </ul>\n";
          }
          $backtrace_text .= "  </li>\n";
        }
        $backtrace_text .= "</ul>\n";
        $form['tests']['result']['backtrace'] = [
          '#type' => 'item',
          '#title' => $this->t('Backtrace:'),
          '#markup' => $backtrace_text,
        ];
      }
    }

    // Actions.
    $form['actions'] ??= [
      '#type' => 'actions',
    ];

    $form['actions']['save'] = [
      '#type' => 'submit',
      '#button_type' => 'primary',
      '#value' => $this->t('Save'),
      '#name' => 'save_config',
      '#validate' => ['::validateSaveForm'],
      '#submit' => ['::submitSaveForm'],
    ];

    $form['actions']['test'] = [
      '#type' => 'submit',
      '#value' => $this->t('Test'),
      '#name' => 'run_tests',
      '#validate' => ['::validateTestForm'],
      '#submit' => ['::submitTestForm'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // We need to clean errors for ajax submissions.
    $form_state->clearErrors();
  }

  /**
   * Validate data locator plugin settings.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function validateSaveForm(array &$form, FormStateInterface $form_state) {
    // Clear required field errors.
    $form_state->clearErrors();
    $this->dataManager->validateDataLocatorConfigurationForm($form, $form_state);
  }

  /**
   * Validate test content.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function validateTestForm(array &$form, FormStateInterface $form_state) {
    // Clear required field errors.
    $form_state->clearErrors();
    $metadata_raw = $form_state->getValue('metadata');
    if (empty($metadata_raw)) {
      $form_state->setErrorByName(
        'metadata',
        $this->t('Metadata field is required.')
      );
      return;
    }
    try {
      $metadata = Yaml::parse($metadata_raw);
      if (!is_array($metadata)) {
        $metadata = [$metadata];
      }
    }
    catch (ParseException $e) {
      $err_line = $e->getParsedLine();
      if (0 <= $err_line) {
        $form_state->setErrorByName(
          'metadata',
          $this->t(
            'Failed to parse metadata (Yaml) near line @line: @error',
            ['@line' => $e->getParsedLine(), '@error' => $e->getSnippet()]
          )
        );
      }
      else {
        $form_state->setErrorByName(
          'metadata',
          $this->t('Failed to parse metadata (Yaml).')
        );
      }
      return;
    }
    $form_state->set('metadata', $metadata);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * Save data locator plugin settings.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function submitSaveForm(array &$form, FormStateInterface $form_state) {
    $this->dataManager->submitDataLocatorConfigurationForm($form, $form_state);
    \Drupal::messenger()->addMessage('Data locator configuration saved.');
  }

  /**
   * Submit tests.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function submitTestForm(array &$form, FormStateInterface $form_state) {
    $metadata = $form_state->get('metadata');
    $provide_default = $form_state->getValue('provide_default', FALSE);
    try {
      $result = $this->dataManager->locateUri(
        $metadata,
        $provide_default,
        $backtrace
      );
    }
    catch (MetadataException $exception) {
      $result = $exception->getMessage();
    }
    $form_state->set('result', $result);
    $form_state->set('backtrace', $backtrace);
    $form_state->setRebuild();
  }

}
