<?php

namespace Drupal\genoring\Form;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\genoring\Event\DataModelManagementEvent;
use Drupal\genoring\Event\GenoringEvents;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class DatasetFormBase.
 *
 * Base form used for creation, edit and delete forms for GenoRing datasets.
 *
 * @ingroup genoring
 */
class DatasetFormBase extends EntityForm {

  /**
   * An entity query factory for the dataset entity type.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $entityStorage;

  /**
   * Construct the DatasetFormBase.
   *
   * @param \Drupal\Core\Entity\EntityStorageInterface $entity_storage
   *   An entity query factory for the dataset entity type.
   */
  public function __construct(EntityStorageInterface $entity_storage) {
    $this->entityStorage = $entity_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $form = new static($container->get('entity_type.manager')->getStorage('genoring_dataset'));
    $form->setMessenger($container->get('messenger'));
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, ?string $model_id = NULL) {
    $form = parent::buildForm($form, $form_state);
    $dataset = $this->getEntity();
    // Disable form cache as we need to recalculate subform contents each time.
    $form['#cache'] = ['max-age' => 0];
    $form['label'] = [
      '#type' => 'textfield',
      // @todo Add a mecanism to change this label according to a selected model name.
      '#title' => $this->t('Dataset name'),
      '#maxlength' => 255,
      '#default_value' => $dataset->label(),
      '#required' => TRUE,
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#title' => $this->t('Machine name'),
      '#default_value' => $dataset->id(),
      '#machine_name' => [
        'exists' => [$this, 'exists'],
        'replace_pattern' => '([^a-z0-9_]+)',
        'error' => 'The machine-readable name must be unique, and can only contain lowercase letters, numbers, and underscores.',
      ],
      '#disabled' => !$dataset->isNew(),
    ];

    if (!empty($model_id)) {
      $form['data_models'] = $this->getDataModelManagementForm([], $form_state, $model_id);
    }
    else {
      // List data files and provide file management interface.
      $form['data_files'] = $this->getFileManagementForm([], $form_state);
      $form['data_models'] = $this->getDataModelManagementForm([], $form_state);
    }

    return $form;
  }

  /**
   * Returns the form title.
   *
   * @return string
   *   The form page title.
   */
  public function getTitle() {
    $model_id = \Drupal::routeMatch()->getParameter('model_id');
    // Get data model definitions.
    $data_manager = \Drupal::service('genoring.data_manager');
    $data_model_definitions = $data_manager->getDataModelDefinitions();
    if (!empty($data_model_definitions[$model_id])) {
      return $this->t('Add @model', ['@model' => $data_model_definitions[$model_id]['label']]);
    }
    return $this->t('Add dataset');
  }

  /**
   * Provides the sub-form to manage dataset files.
   *
   * @param array $form
   *   A default form that could be empty.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The dataset file management form.
   */
  public function getFileManagementForm(
    array $form,
    FormStateInterface $form_state,
  ) :array {
    $dataset = $this->getEntity();
    $files = $dataset->getDataFiles();
    if ($dataset->isNew() && empty($files)) {
      $form += [
        '#type' => 'container',
      ];
      return $form;
    }

    $form += [
      '#type' => 'details',
      '#title' => $this->t('Data files'),
      '#open' => FALSE,
      '#tree' => TRUE,
    ];

    // List files.
    foreach ($dataset->getDataFiles() as $file_index => $data_file) {
      $form[$file_index] = [
        '#type' => 'container',
        '#attributes' => ['class' => ['container-inline']],
      ];
      $form[$file_index]['data_file'] = [
        '#type' => 'item',
        '#markup' => $this->t(
          'File: @filename',
          ['@filename' => $data_file->getFilename()],
        ),
      ];
      // @code
      //   $form[$file_index]['replace_file'] = [
      //     '#type' => 'submit',
      //     '#name' => 'replace_file',
      //     '#limit_validation_errors' => [],
      //     '#submit' => ['::submitReplaceFile'],
      //     '#value' => $this->t('Replace'),
      //   ];
      //   $form[$file_index]['remove_file'] = [
      //     '#type' => 'submit',
      //     '#name' => 'remove_file',
      //     '#limit_validation_errors' => [],
      //     '#submit' => ['::submitRemoveFile'],
      //     '#value' => $this->t('Remove'),
      //   ];
      // @endcode
      // @todo Add file metadata management.
      // @todo Add file operations.
    }
    // Add new file.
    $form['add_file'] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];
    $form['add_file']['new_file'] = [
      '#type' => 'genoring_input',
      '#title' => $this->t('New data file'),
      '#open' => FALSE,
    ];
    $form['add_file']['add_file_button'] = [
      '#type' => 'submit',
      '#limit_validation_errors' => [['data_files', 'add_file', 'new_file']],
      '#validate' => ['::validateAddFile'],
      '#submit' => ['::submitAddFile'],
      '#value' => $this->t('Add'),
    ];
    return $form;
  }

  /**
   * Provides the sub-form to manage dataset data models.
   *
   * @param array $form
   *   A default form that could be empty.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param string $model_id
   *   For a new dataset, it will only display the dataset type form
   *   corresponding to the given model. For an existing dataset, this parameter
   *   is ignored.
   *
   * @return array
   *   The dataset data model management form.
   */
  public function getDataModelManagementForm(
    array $form,
    FormStateInterface $form_state,
    string $model_id = '',
  ) :array {
    $data_manager = \Drupal::service('genoring.data_manager');
    $dataset = $this->getEntity();
    $current_data_models = $dataset->getDataModelInstances();
    // Get list of available data models.
    $data_model_definitions = $data_manager->getDataModelDefinitions();

    if (empty($current_data_models)) {
      $form += [
        '#type' => 'container',
        '#tree' => TRUE,
      ];
    }
    else {
      $form += [
        '#type' => 'details',
        '#title' => $this->t('Data models'),
        '#open' => TRUE,
        '#tree' => TRUE,
      ];
      // @todo Add dataset settings such as privacy status, etc.
      // Add data model management.
      // - List of managed data models (with used data files)
      //   Trigger data model management event with used model and data files to
      //   get a management form (action buttons with own validate/submit
      //   actions, reports, last updates, etc.) and a "remove data model"
      //   button.
      foreach ($current_data_models as $existing_model_id => $model) {
        $form[$existing_model_id] = [
          '#type' => 'fieldset',
          '#title' => $data_model_definitions[$existing_model_id]['label'] ?? $model->getLabel() ?: '',
          '#description' => $data_model_definitions[$existing_model_id]['description'] ?? $model->getDescription() ?: '',
        ];
        $event = new DataModelManagementEvent($model);
        \Drupal::service('event_dispatcher')->dispatch(
          $event,
          GenoringEvents::DATA_MODEL_MANAGEMENT
        );
        $management_form = $event->getManagementForm();
        if (!empty($management_form)) {
          $form[$existing_model_id] = NestedArray::mergeDeep(
            $management_form,
            $form[$existing_model_id]
          );
        }
        else {
          $form[$existing_model_id] += [
            'info' => [
              '#type' => 'item',
              '#markup' => $this->t('INFO: This data model has currently no operations that could be applied.'),
            ],
          ];
        }
        // Allow alteration.
        if (!empty($data_model_definitions[$existing_model_id]['form']['alter'])) {
          $data_model_definitions[$existing_model_id]['form']['alter']($form, $form_state);
        }
        // @todo To complete.
      }
    }

    // Add data model.
    $form['add_data_model'] = [
      '#type' => 'container',
    ];
    $data_model_details_id = 'data-model-details-wrapper';
    // Check if we got a specific model to create.
    if (!empty($model_id)) {
      // Make sure the model exists.
      if (empty($data_model_definitions[$model_id])) {
        throw new NotFoundHttpException();
      }
      $form['add_data_model']['data_model'] = [
        '#type' => 'hidden',
        '#value' => $model_id,
      ];
      $form_state->set('models.' . $model_id, $data_model_definitions[$model_id]);
    }
    else {
      $form['add_data_model'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Add new data model instance'),
      ]
      + $form['add_data_model'];

      // Otherwise, provide a model selection drowpdown.
      $data_models = [
        '' => '-- Select --',
      ];
      foreach ($data_model_definitions as $available_data_model_id => $data_model_definition) {
        // Only provide non-instanciated data models.
        if (empty($current_data_models[$available_data_model_id])) {
          $data_models[$available_data_model_id] = $data_model_definition['label'];
        }
      }
      // Get selected model from state if set, or from value otherwise.
      $model_id =
        $form_state->get('data_model')
        ?? $form_state->getValue(
          ['data_models', 'add_data_model', 'data_model'],
          ''
        );
      if (1 < count($data_models)) {
        $form['add_data_model']['data_model'] = [
          '#type' => 'select',
          '#title' => $this->t('Data Model'),
          '#default_value' => $model_id,
          '#options' => $data_models,
          '#ajax' => [
            'callback' => [$this, 'buildAjaxDataModelForm'],
            'wrapper' => $data_model_details_id,
            'method' => 'replaceWith',
            'effect' => 'fade',
          ],
        ];
      }
      else {
        // No data model to add, empty section.
        $form['add_data_model'] = [
          '#type' => 'container',
        ];
      }
    }

    if (empty($data_model_definitions[$model_id])) {
      $form['add_data_model']['data_model_details'] = [
        '#type' => 'container',
        '#attributes' => [
          'id' => $data_model_details_id,
        ],
      ];
    }
    else {
      $form_state->set('models.' . $model_id, $data_model_definitions[$model_id]);
      $data_model = $data_model_definitions[$model_id];
      $form['add_data_model']['data_model_details'] = [
        '#type' => 'details',
        '#title' => $this->t('@model', ['@model' => $data_model['label']]),
        '#open' => TRUE,
        'description' => [
          '#type' => 'item',
          '#markup' => $data_model['description'],
        ],
        '#attributes' => [
          'id' => $data_model_details_id,
        ],
      ];
      foreach ($data_model['dataset_fields'] as $field_name => $field_info) {
        $form['add_data_model']['data_model_details'][$field_name] = $data_manager->generateFormField($field_name, $field_info);
      }
    }
    return $form;
  }

  /**
   * Handles switching the selected field mapper plugin.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxDataModelForm(array $form, FormStateInterface $form_state) {
    return $form['data_models']['add_data_model']['data_model_details'] ?? [];
  }

  /**
   * Checks for an existing dataset.
   *
   * @param string|int $entity_id
   *   The entity ID.
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return bool
   *   TRUE if this format already exists, FALSE otherwise.
   */
  public function exists($entity_id, array $element, FormStateInterface $form_state) {
    $query = $this->entityStorage->getQuery();
    $result = $query->condition('id', $element['#field_prefix'] . $entity_id)
      ->accessCheck()
      ->execute();
    return (bool) $result;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Save');
    return $actions;
  }

  /**
   * {@inheritdoc}
   */
  public function validateAddFile(array &$form, FormStateInterface $form_state) {
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // @todo Check for file change/removal if the file is in use by a data
    // model.

    // Keep track of selected data model to have the correct form generated
    // between ajax calls (eg. file uploads).
    $model_id = $form_state->getValue(['data_models', 'add_data_model', 'data_model'], '');
    $form_state->set('data_model', $model_id);

    // Manage data model validations.
    if (!empty($model_id)) {
      $model_definition = $form_state->get('models.' . $model_id);
      if (!empty($model_definition['form']['validate'])) {
        $model_definition['form']['validate']($form, $form_state);
      }
    }

    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $dataset = $this->getEntity();
    $file_uris = $dataset->getDataFileUris();

    // @todo When a second file is added, the form is saved while it should not.
    // Prevent saving when the 'save' button has not been used.

    // Manage data models.
    if (empty($model_id)) {
      $model_id =
        $form_state->get('data_model')
        ?? $form_state->getValue(['data_models', 'add_data_model', 'data_model']);
    }
    if (!empty($model_id)) {
      // Add the new selected data model.
      $subform_state = GenoringSubformState::createForSubform(
        ['data_models', 'add_data_model', 'data_model_details'],
        $form,
        $form_state
      );

      $this->addDataModel($form['data_models']['add_data_model']['data_model_details'], $subform_state, $model_id);
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitReplaceFile(array &$form, FormStateInterface $form_state) {
    if (($trigger = $form_state->getTriggeringElement())
        && ($file_index = $trigger['#parents'][1] ?? '')
    ) {
      $dataset = $this->getEntity();
      $file_uris = $dataset->getDataFileUris();
      $replaced_file = $file_uris[$file_index];
      // @todo Manage file replacement.
      $this->messenger()->addWarning('Not implemented.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitRemoveFile(array &$form, FormStateInterface $form_state) {
    if (($trigger = $form_state->getTriggeringElement())
        && ($file_index = $trigger['#parents'][1] ?? '')
    ) {
      $dataset = $this->getEntity();
      $file_uris = $dataset->getDataFileUris();
      $removed_file = $file_uris[$file_index];

      // Keep only file name.
      if ($slash_pos = strrpos($removed_file, '/')) {
        $removed_file = substr($removed_file, $slash_pos + 1);
      }
      unset($file_uris[$file_index]);
      $dataset
        ->setDataFileUris(array_values($file_uris))
        ->save();
      $this->messenger()->addMessage('File ' . $removed_file . ' has been removed from the dataset successfully.');
      // @todo The file is removed from the dataset but is still managed by
      // Drupal. That file may or may not be in use by another dataset. We
      // should provide a separate interface to manage data files.
      $form_state->setRebuild(TRUE);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitAddFile(array $form, FormStateInterface $form_state) {
    $dataset = $this->getEntity();
    $new_file_input = $form_state->getValue(['data_files', 'add_file', 'new_file'], []);
    $data_manager = \Drupal::service('genoring.data_manager');
    $new_files = $data_manager->saveGenoringInput($new_file_input);
    if (!empty($new_files)) {
      foreach ($new_files as $new_file) {
        $dataset->addDataFile($new_file);
      }
      $dataset->save();
      $this->messenger()->addMessage(
        t('Successfully added file to dataset.')
      );
    }
    $form_state->setRedirect('entity.genoring_dataset.edit_form', ['genoring_dataset' => $dataset->id()]);
  }

  /**
   * Add a new data model to current dataset.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param string $model_id
   *   Identifier of data model to add.
   */
  public function addDataModel(array $form, FormStateInterface $form_state, string $model_id) {
    $dataset = $this->getEntity();
    $entity_type_manager = \Drupal::entityTypeManager();
    $entity_display_repository = \Drupal::service('entity_display.repository');
    $data_manager = \Drupal::service('genoring.data_manager');
    // Create a new external entity data model.
    $model_definition = $form_state->get('models.' . $model_id);
    $xntt_type_id = $dataset->id() . '__' . $model_id;
    // Check if the external entiy type already exists.
    $xntt = $entity_type_manager->getStorage('external_entity_type')->load($xntt_type_id);
    if (empty($xntt)) {
      // Not existing, create one.
      // Prepare storage clients config.
      $data_aggregator_config = [
        'id' => 'model',
        'config' => [
          'storage_clients' => [],
        ],
      ];
      foreach ($model_definition['sources'] as $source_id => $source_clients) {
        foreach ($source_clients as $source_client) {
          $data_aggregator_config['config']['storage_clients'][] =
            $data_manager->generateStorageClientConfig(
              $source_client['id'],
              $source_client['config'] ?? [],
              $source_client['aggr'] ?? []
            );
        }
      }
      // Create a new external entity type as data model.
      $xntt = $entity_type_manager->getStorage('external_entity_type')->create([
        'id' => $xntt_type_id,
        'label' => $dataset->label() . ' ' . $model_definition['label'],
        'label_plural' => $dataset->label() . ' ' . $model_definition['label'],
        'description' => '',
        'generate_aliases' => FALSE,
        'read_only' => FALSE,
        'debug_level' => 0,
        // Field mappers are added after fields.
        'field_mappers' => [],
        'data_aggregator' => $data_aggregator_config,
        'persistent_cache_max_age' => 0,
      ]);
      $xntt->save();
    }
    else {
      // The associated external entity exists, complete/update.
      // @todo Update existing model.
      $this->messenger()->addWarning(
        $this->t(
          'WARNING: the model @model already exists.',
          [
            '@model' => $xntt_type_id,
          ]
        )
      );
    }
    // Manage fields.
    foreach ($model_definition['model_fields'] as $field_name => $field_info) {
      try {
        // Make sure we work on a non-base field.
        if (!empty($field_info['type'])) {
          // Storage.
          $field_storage_config = $entity_type_manager
            ->getStorage('field_storage_config')
            ->load($xntt_type_id . '.' . $field_name);
          if (empty($field_storage_config)) {
            // Add missing field storage.
            $field_storage_array = [
              'field_name' => $field_name,
              'entity_type' => $xntt_type_id,
            ]
            + $data_manager->generateFieldStorage(
              $field_name,
              $field_info
            );
            $field_storage_config = $entity_type_manager
              ->getStorage('field_storage_config')
              ->create($field_storage_array);
            $field_storage_config->save();
          }
          // Field config.
          $field_config_id =
            $xntt_type_id
            . '.'
            . $xntt_type_id
            . '.'
            . $field_name;
          $field_config = $entity_type_manager
            ->getStorage('field_config')
            ->load($field_config_id);
          if (empty($field_config)) {
            // Add missing field config.
            $field_config_array = [
              'field_name' => $field_name,
              'entity_type' => $xntt_type_id,
              'bundle' => $xntt_type_id,
            ]
            + $data_manager->generateFieldConfig(
              $field_name,
              $field_info
            )
            + [
              'label' => ucfirst($field_name),
              'description' => '',
            ];
            $entity_type_manager
              ->getStorage('field_config')
              ->create($field_config_array)
              ->save();
          }
          // Entity form display.
          $field_form_display_array = $data_manager->generateFieldFormDisplay(
            $field_name,
            $field_info
          );
          if (!empty($field_form_display_array)) {
            $entity_display_repository
              ->getFormDisplay($xntt_type_id, $xntt_type_id, 'default')
              ->setComponent($field_name, $field_form_display_array)
              ->save();
          }
          // Entity view display.
          $field_view_display_array = $data_manager->generateFieldViewDisplay(
            $field_name,
            $field_info
          );
          if (!empty($field_view_display_array)) {
            $entity_display_repository
              ->getViewDisplay($xntt_type_id, $xntt_type_id)
              ->setComponent($field_name, $field_view_display_array)
              ->save();
          }
        }
      }
      catch (\Exception $e) {
        $error = TRUE;
        $this->messenger()->addError(
          $this->t(
            'There was a problem managing field %label: @message',
            ['%label' => $field_name, '@message' => $e->getMessage()]
          )
        );
      }
    }
    // Since we added new fields, we need to clear new external entity field
    // cache to define their mapping.
    $xntt->getMappableFields(TRUE);
    // Mapping definition.
    foreach ($model_definition['model_fields'] as $field_name => $field_info) {
      if (!empty($field_info['mapping'])) {
        foreach ($field_info['mapping'] as $source_id => $source_mapping) {
          $field_mapping = $data_manager->generateFieldMapping(
            $source_mapping['mapping'],
            $source_mapping['data_processors'] ?? [],
            $source_mapping['overrides'] ?? []
          );
          $xntt
            ->setFieldMapperId($field_name, $field_mapping['id'])
            ->setFieldMapperConfig($field_name, $field_mapping['config']);
        }
      }
    }
    // Save new mapping.
    $xntt->save();

    // @todo Lock xntt parts that should not be changed.

    // Add model to dataset.
    $dataset->addDataModel($model_id);
    $dataset->save();
    $this->messenger()->addMessage('Data model ' . $model_id . ' for dataset ' . $dataset->id() . ' has been created.');
    if (!empty($model_definition['form']['submit'])) {
      // @todo create substate $form_state
      $model_definition['form']['submit']($form, $form_state, $model_id, $dataset, $xntt);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $dataset = $this->getEntity();

    $status = $dataset->save();

    $url = $dataset->toUrl();

    $edit_link = Link::fromTextAndUrl($this->t('Edit'), $url)->toString();

    if ($status == SAVED_UPDATED) {
      $this->messenger()->addMessage(
        $this->t('Dataset %label has been updated.',
        ['%label' => $dataset->label()])
      );
      $this->logger('genoring')->notice(
        'Dataset %label has been updated.',
        ['%label' => $dataset->label(), 'link' => $edit_link]
      );
    }
    else {
      $this->messenger()->addMessage(
        $this->t('Dataset %label has been added.',
        ['%label' => $dataset->label()])
      );
      $this->logger('genoring')->notice(
        'Dataset %label has been added.',
        ['%label' => $dataset->label(), 'link' => $edit_link]
      );
    }

    $form_state->setRedirect('entity.genoring_dataset.edit_form', ['genoring_dataset' => $dataset->id()]);
    return $status;
  }

}
