<?php

namespace Drupal\genoring\Plugin\Genoring\DataLocator;

use Drupal\Core\Form\FormStateInterface;
use Drupal\genoring\DataLocator\DataLocatorBase;
use JsonPath\InvalidJsonException;
use JsonPath\InvalidJsonPathException;
use JsonPath\JsonObject;

/**
 * Generated file locator.
 *
 * @DataLocator(
 *   id = "term",
 *   label = @Translation("Term locator"),
 *   description = @Translation("Data locator used to place files according to its metadata terms."),
 *   defaultSupportedMetadata = {}
 * )
 */
class MetadataTermLocator extends DataLocatorBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'data_uri' => '',
      'term' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state,
  ) {
    $configuration = $this->getConfiguration();
    $form['term'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Term to use:'),
      '#description' => $this->t('Leave empty to only use the specified parent directory. Start by a "$" to specify a JSON path, otherwise just use the term name. If using JSON Path, only non-array values returned by the expression will be taken into account.'),
      '#size' => 40,
      '#wrapper_attributes' => ['class' => ['container-inline']],
      '#required' => FALSE,
      '#default_value' => $form_state->getValue(
        'term',
        $configuration['term']
        ?? ''
        ?: static::defaultConfiguration()['term']
      ),
    ];
    $form['data_uri'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Parent directory:'),
      '#description' => $this->t('Parent directory to store files in sub-directories using metadata terms. This can be empty.'),
      '#required' => FALSE,
      '#size' => 40,
      '#wrapper_attributes' => ['class' => ['container-inline']],
      '#default_value' => $form_state->getValue(
        'data_uri',
        $configuration['data_uri']
        ?? ''
        ?: static::defaultConfiguration()['data_uri']
      ),
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function locateUri(
    array $metadata,
    string $base_uri = '',
  ) :string {
    $uri = $base_uri;
    [$stream, $path, $filename] = $this->explodeUri($uri);
    $configuration = $this->getConfiguration();
    // Get term.
    $term = '';
    if (!empty($configuration['term'])) {
      if ('$' === $configuration['term'][0]) {
        // JSON Path.
        try {
          $json_object = new JsonObject($metadata, TRUE);
          $result = $json_object->get($configuration['term']);
          if (!empty($result) && !is_array($result)) {
            $term = $result;
          }
        }
        catch (InvalidJsonException | InvalidJsonPathException $exception) {
          $this->logger->error(
            'Failed to parse JSON Path ('
            . $configuration['term']
            . '): '
            . $exception
          );
        }
      }
      else {
        // Direct term.
        $term = $metadata[$configuration['term']] ?? '';
      }
    }
    // Normalize term.
    if (!is_string($term)) {
      $term = '';
    }
    $term = strtolower(
      trim(
        preg_replace('/\W/', '_', $term),
        '_'
      )
    );
    if (!empty($term)) {
      $term .= '/';
    }
    // Prepend parent directory if one.
    if (!empty($configuration['data_uri'])) {
      $term = trim($configuration['data_uri'], '/') . '/' . $term;
    }
    // Apply changes.
    if (!empty($metadata['type'])) {
      if ('file' === $metadata['type']) {
        $uri = $stream . $path . $term . $filename;
      }
      elseif ('directory' === $metadata['type']) {
        $uri = $stream . $path . $term;
      }
    }
    return $uri;
  }

}
