<?php

namespace Drupal\Tests\genoring\Unit;

use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\genoring\DataManager\DataManager;
use Drupal\genoring\Exception\DataLocatorException;
use Drupal\genoring\Plugin\Genoring\DataLocator\DefaultLocator;

/**
 * The class to test default data locator.
 *
 * @group genoring
 */
class DataLocatorDefaultUnitTest extends UnitTestCase {

  /**
   * Data provider for testExplodeUri().
   *
   * Structure:
   * - expected
   * - uri
   * - allow_no_stream
   * - test name.
   */
  public static function provideTestExplodeUri() {
    return [
      [
        ['public://', 'somedir/some_sub_dir/', 'somefile.txt'],
        'public://somedir/some_sub_dir/somefile.txt',
        FALSE,
        'A full URI',
      ],
      [
        ['private://', 'somedir/', 'somefile.txt'],
        'private://somedir/somefile.txt',
        FALSE,
        'A private full URI',
      ],
      [
        ['private://', 'somedir/', ''],
        'private://somedir/',
        FALSE,
        'URI with no file',
      ],
      [
        ['public://', 'somedir/', ''],
        'somedir/',
        FALSE,
        'URI with no scheme',
      ],
      [
        [NULL, 'somedir/', ''],
        'somedir/',
        TRUE,
        'URI with no scheme to return',
      ],
      [
        [NULL, '', ''],
        '',
        TRUE,
        'Empty URI',
      ],
      [
        ['public://', '', ''],
        '',
        FALSE,
        'Empty URI but force scheme',
      ],
    ];
  }

  /**
   * Tests DataLocatorBase::explodeUri().
   *
   * @param array $expected
   *   Expected result.
   * @param string $uri
   *   Input URI.
   * @param bool $allow_no_stream
   *   Allow or not to return no stream if not detected.
   * @param string $test_name
   *   Test name.
   *
   * @dataProvider provideTestExplodeUri
   */
  public function testExplodeUri(array $expected, string $uri, bool $allow_no_stream, string $test_name) {
    $string_translation = $this->createMock(TranslationInterface::class);

    $logger = $this->createMock(LoggerChannelInterface::class);
    $logger_factory = $this->createMock(LoggerChannelFactoryInterface::class);
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('genoring_locator_default')
      ->willReturn($logger);

    $data_locator = new DefaultLocator([], 'default', [], $string_translation, $logger_factory);

    $reflectionMethod = new \ReflectionMethod($data_locator, 'explodeUri');
    $reflectionMethod->setAccessible(TRUE);

    $result = $reflectionMethod->invoke($data_locator, $uri, $allow_no_stream);
    $this->assertEquals($expected, $result, $test_name);
  }

  /**
   * Data provider for testExplodeUriError().
   *
   * Structure:
   * - uri
   * - test name.
   */
  public static function provideTestExplodeUriError() {
    return [
      [
        'public:////somedir/some sub dir/somefile.txt',
        'Invalid URI 1',
      ],
      [
        'public://somedir/private://someother/somefile.txt',
        'Invalid URI 2',
      ],
      [
        'public/private://someother/somefile.txt',
        'Invalid URI 3',
      ],
      [
        'pub-priv://someother/somefile.txt',
        'Invalid URI 4',
      ],
    ];
  }

  /**
   * Tests errors for DataLocatorBase::explodeUri().
   *
   * @param string $uri
   *   Input invalid URI.
   * @param string $test_name
   *   Test name.
   *
   * @dataProvider provideTestExplodeUriError
   */
  public function testExplodeUriError(string $uri, string $test_name) {
    $string_translation = $this->createMock(TranslationInterface::class);

    $logger = $this->createMock(LoggerChannelInterface::class);
    $logger_factory = $this->createMock(LoggerChannelFactoryInterface::class);
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('genoring_locator_default')
      ->willReturn($logger);

    $data_locator = new DefaultLocator([], 'default', [], $string_translation, $logger_factory);

    $reflectionMethod = new \ReflectionMethod($data_locator, 'explodeUri');
    $reflectionMethod->setAccessible(TRUE);

    $this->expectException(DataLocatorException::class);
    $this->expectExceptionMessage("Invalid URI: '$uri'");

    $result = $reflectionMethod->invoke($data_locator, $uri);
  }

  /**
   * Data provider for testStringCaseProcessor().
   *
   * Structure:
   * - expected_uri
   * - base_uri
   * - metadata
   * - config
   * - test name.
   */
  public static function provideTestDefaultDataLocator() {
    return [
      [
        DataManager::$defaultConfiguration['data_uri'] . '/',
        '',
        [],
        [],
        'Default base URI',
      ],
      [
        DataManager::$defaultConfiguration['data_uri'] . '/somedir/',
        'somedir/',
        [],
        [],
        'Default base URI given a subdirectory',
      ],
      [
        DataManager::$defaultConfiguration['data_uri'] . '/file.txt',
        'file.txt',
        [],
        [],
        'Default base URI given a file',
      ],
      [
        DataManager::$defaultConfiguration['data_uri'] . '/somedir/file.txt',
        'somedir/file.txt',
        [],
        [],
        'Default base URI given a subdirectory with file',
      ],
      [
        'public://geno/data/somedir/',
        'public://somedir/',
        [],
        [
          'data_uri' => 'private://geno/data/',
          'replace' => 0,
        ],
        'A custom base with scheme, no replacement, given a directory with scheme',
      ],
      [
        'public://geno/data/somedir/somefile.txt',
        'public://somedir/somefile.txt',
        [],
        [
          'data_uri' => 'private://geno/data/',
          'replace' => 0,
        ],
        'A custom base with scheme, no replacement, given a directory with scheme and file',
      ],
      [
        'private://geno/data/somedir/somefile.txt',
        'somedir/somefile.txt',
        [],
        [
          'data_uri' => 'private://geno/data/',
          'replace' => 0,
        ],
        'A custom base with scheme, no replacement, given a directory and file with no scheme',
      ],
      [
        'private://geno/data/somedir/somefile.txt',
        '/somedir/somefile.txt',
        [],
        [
          'data_uri' => 'private://geno/data/',
          'replace' => 0,
        ],
        'A custom base with scheme, no replacement, given an absolute directory and file with no scheme',
      ],
      [
        'private://geno/data/somedir/',
        '/somedir/',
        [],
        [
          'data_uri' => '/geno/data/',
          'replace' => 0,
        ],
        'A custom base with no scheme, no replacement, given a directory and file with no scheme',
      ],
      [
        'private://geno/data/somedir/',
        '//somedir/',
        [],
        [
          'data_uri' => '/geno/data/',
          'replace' => 0,
        ],
        'A custom base with no scheme, no replacement, given an absolute directory with no stream',
      ],
      [
        'private://geno/data/somedir/',
        '://somedir/',
        [],
        [
          'data_uri' => '/geno/data/',
          'replace' => 0,
        ],
        'A custom base with scheme, no replacement, given an absolute directory with incomplete stream',
      ],
      [
        'private://custom/',
        'public://somedir/',
        [],
        [
          'data_uri' => 'private://custom',
          'replace' => 1,
        ],
        'A custom base with scheme, with replacement, given a directory and stream',
      ],
      [
        'private://custom/somefile.txt',
        'public://somedir/somefile.txt',
        [],
        [
          'data_uri' => 'private://custom/',
          'replace' => 1,
        ],
        'A custom base directory that should replace erything but keep file name.',
      ],
      [
        'private://somefile.txt',
        'public://somedir/somefile.txt',
        [],
        [
          'data_uri' => 'private://',
          'replace' => 1,
        ],
        'A scheme change.',
      ],
    ];
  }

  /**
   * Tests default data locator.
   *
   * @dataProvider provideTestDefaultDataLocator
   */
  public function testDefaultDataLocator(
    string $expected_uri,
    string $base_uri,
    array $metadata,
    array $config,
    string $test_name,
  ) {
    $string_translation = $this->createMock(TranslationInterface::class);

    $logger = $this->createMock(LoggerChannelInterface::class);
    $logger_factory = $this->createMock(LoggerChannelFactoryInterface::class);
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('genoring_locator_default')
      ->willReturn($logger);

    $data_locator = new DefaultLocator($config, 'default', [], $string_translation, $logger_factory);
    $result_uri = $data_locator->locateUri($metadata, $base_uri);
    $this->assertEquals($expected_uri, $result_uri, $test_name);
  }

}
