<?php

namespace Drupal\Tests\genoring\Unit;

use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\genoring\Plugin\Genoring\DataLocator\TaxonomyLocator;
use Drupal\genoring\Taxonomy\TaxonomyInterface;

/**
 * The class to test taxonomy data locator.
 *
 * @group genoring
 */
class DataLocatorTaxonomyUnitTest extends UnitTestCase {

  /**
   * Test container.
   *
   * @var \Symfony\Component\DependencyInjection\ContainerInterface
   */
  protected $container;

  /**
   * Mocked taxonomy serice.
   *
   * @var \SDrupal\genoring\Taxonomy\TaxonomyInterface
   */
  protected $taxonomyServiceMock;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $container = new ContainerBuilder();

    $taxonomy_lookup = [
      'taxid' => [
        3702                   => 3702,
        'ARATH'                => 3702,
        'arabidopsis_thaliana' => 3702,
        'Arabidopsis thaliana' => 3702,
        4641                   => 4641,
        'MUSAC'                => 4641,
        'musa_acuminata'       => 4641,
        'Musa acuminata'       => 4641,
      ],
      'species_codes' => [
        3702                   => 'ARATH',
        'ARATH'                => 'ARATH',
        'arabidopsis_thaliana' => 'ARATH',
        'Arabidopsis thaliana' => 'ARATH',
        4641                   => 'MUSAC',
        'MUSAC'                => 'MUSAC',
        'musa_acuminata'       => 'MUSAC',
        'Musa acuminata'       => 'MUSAC',
      ],
      'species_names' => [
        3702                   => 'arabidopsis_thaliana',
        'ARATH'                => 'arabidopsis_thaliana',
        'arabidopsis_thaliana' => 'arabidopsis_thaliana',
        'Arabidopsis thaliana' => 'arabidopsis_thaliana',
        4641                   => 'musa_acuminata',
        'MUSAC'                => 'musa_acuminata',
        'musa_acuminata'       => 'musa_acuminata',
        'Musa acuminata'       => 'musa_acuminata',
      ],
    ];

    // Create a mock for the taxonomy service.
    $this->taxonomyServiceMock = $this->createMock(TaxonomyInterface::class);
    $this->taxonomyServiceMock->method('getSpeciesCode')
      ->willReturnCallback(function (string $string) use ($taxonomy_lookup) {
        return $taxonomy_lookup['species_codes'][$string] ?? TaxonomyInterface::SPECIES_CODE_UNKNOWN;
      });
    $this->taxonomyServiceMock->method('getTaxonomyId')
      ->willReturnCallback(function (string $string) use ($taxonomy_lookup) {
        return $taxonomy_lookup['taxid'][$string] ?? TaxonomyInterface::TAXONOMY_ID_UNKNOWN;
      });
    $this->taxonomyServiceMock->method('getNormalizedSpeciesName')
      ->willReturnCallback(function (string $string) use ($taxonomy_lookup) {
        return $taxonomy_lookup['species_names'][$string] ?? TaxonomyInterface::SPECIES_NAME_UNKNOWN;
      });
    $container->set('genoring.taxonomy', $this->taxonomyServiceMock);

    \Drupal::setContainer($container);
    $this->container = $container;
  }

  /**
   * Data provider for testStringCaseProcessor().
   *
   * Structure:
   * - expected_uri
   * - base_uri
   * - metadata
   * - config
   * - test name.
   */
  public static function provideTestTaxonomyDataLocator() {
    return [
      [
        'private://genoring/somebase/',
        'private://genoring/somebase/',
        [],
        [],
        'No type, no taxonomy, no config',
      ],
      [
        'private://genoring/somebase/' . TaxonomyInterface::SPECIES_CODE_UNKNOWN . '/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => '',
        ],
        ['mode' => 'code'],
        'File type, no taxonomy, code',
      ],
      [
        'private://genoring/somebase/subbase/' . TaxonomyInterface::SPECIES_CODE_UNKNOWN . '/some.fasta',
        'private://genoring/somebase/subbase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Dahu vulgaris',
        ],
        ['mode' => 'code'],
        'File type, unknown taxonomy, code',
      ],
      [
        'private://genoring/somebase/subbase/' . TaxonomyInterface::TAXONOMY_ID_UNKNOWN . '/some.fasta',
        'private://genoring/somebase/subbase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Dahu vulgaris',
        ],
        ['mode' => 'taxid'],
        'File type, unknown taxonomy, taxid',
      ],
      [
        'private://genoring/somebase/subbase/' . TaxonomyInterface::SPECIES_NAME_UNKNOWN . '/some.fasta',
        'private://genoring/somebase/subbase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Dahu vulgaris',
        ],
        ['mode' => 'name'],
        'File type, unknown taxonomy, name',
      ],
      [
        'private://genoring/somebase/MUSAC/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'code'],
        'File type, with taxonomy, code',
      ],
      [
        'private://genoring/somebase/plants/ARATH/some.fasta',
        'private://genoring/somebase/plants/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Arabidopsis thaliana',
        ],
        ['mode' => 'code'],
        'File type, with taxonomy, code 2',
      ],
      [
        'private://genoring/somebase/MUSAC/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'MUSAC',
        ],
        ['mode' => 'code'],
        'File type, with taxonomy code, code',
      ],
      [
        'private://genoring/somebase/MUSAC/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 4641,
        ],
        ['mode' => 'code'],
        'File type, with taxonomy id, code',
      ],
      [
        'private://genoring/somebase/MUSAC/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'musa_acuminata',
        ],
        ['mode' => 'code'],
        'File type, with taxonomy normalized name, code',
      ],
      [
        'private://genoring/somebase/MUSAC/',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'directory',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'code'],
        'Directory type, with taxonomy, code',
      ],
      [
        'private://genoring/somebase/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'xxx',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'code'],
        'Unkown type, with taxonomy, code',
      ],
      [
        'private://genoring/somebase/4641/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'taxid'],
        'File type, with taxonomy, taxid',
      ],
      [
        'private://genoring/somebase/4641/',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'directory',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'taxid'],
        'Directory type, with taxonomy, taxid',
      ],
      [
        'private://genoring/somebase/musa_acuminata/some.fasta',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'name'],
        'File type, with taxonomy, name',
      ],
      [
        'private://genoring/somebase/musa_acuminata/',
        'private://genoring/somebase/some.fasta',
        [
          'type' => 'directory',
          'taxonomy' => 'Musa acuminata',
        ],
        ['mode' => 'name'],
        'Directory type, with taxonomy, name',
      ],
      [
        'private://genoring/somebase/plants/' . TaxonomyInterface::SPECIES_CODE_MULTI . '/some.fasta',
        'private://genoring/somebase/plants/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => ['Arabidopsis thaliana', 'Musa acuminata'],
        ],
        ['mode' => 'code'],
        'File type, with multi taxonomies, code',
      ],
      [
        'private://genoring/somebase/plants/' . TaxonomyInterface::TAXONOMY_ID_MULTI . '/some.fasta',
        'private://genoring/somebase/plants/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => ['Arabidopsis thaliana', 'Musa acuminata'],
        ],
        ['mode' => 'taxid'],
        'File type, with multi taxonomies, taxid',
      ],
      [
        'private://genoring/somebase/plants/' . TaxonomyInterface::SPECIES_NAME_MULTI . '/some.fasta',
        'private://genoring/somebase/plants/some.fasta',
        [
          'type' => 'file',
          'taxonomy' => ['Arabidopsis thaliana', 'Musa acuminata', 'Dahu vulgaris'],
        ],
        ['mode' => 'name'],
        'File type, with multi taxonomies, name',
      ],
    ];
  }

  /**
   * Tests Taxonomy data locator.
   *
   * @dataProvider provideTestTaxonomyDataLocator
   */
  public function testTaxonomyDataLocator(
    string $expected_uri,
    string $base_uri,
    array $metadata,
    array $config,
    string $test_name,
  ) {
    $string_translation = $this->createMock(TranslationInterface::class);

    $logger = $this->createMock(LoggerChannelInterface::class);
    $logger_factory = $this->createMock(LoggerChannelFactoryInterface::class);
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('genoring_locator_taxonomy')
      ->willReturn($logger);

    $data_locator = new TaxonomyLocator($config, 'taxonomy', [], $string_translation, $logger_factory);
    $result_uri = $data_locator->locateUri($metadata, $base_uri);
    $this->assertEquals($expected_uri, $result_uri, $test_name);
  }

}
