<?php

namespace Drupal\Tests\genoring\Unit;

use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\genoring\Plugin\Genoring\DataLocator\MetadataTermLocator;

/**
 * The class to test taxonomy data locator.
 *
 * @group genoring
 */
class DataLocatorTermUnitTest extends UnitTestCase {

  /**
   * Data provider for testStringCaseProcessor().
   *
   * Structure:
   * - expected_uri
   * - base_uri
   * - metadata
   * - config
   * - test name.
   */
  public static function provideTestTermDataLocator() {
    return [
      [
        'private://genoring/somebase/',
        'private://genoring/somebase/',
        [],
        [],
        'No type, no term, no config',
      ],
      [
        'private://genoring/somebase/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
        ],
        [],
        'File type, no term, no config',
      ],
      [
        'private://genoring/somebase/',
        'private://genoring/somebase/',
        [],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'No type, no metadata, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => 'bla bla!',
        ],
        [
          'term' => '',
          'data_uri' => 'sorted',
        ],
        'File type, with term, with config data_uri and no term',
      ],
      [
        'private://genoring/somebase/bla_bla/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => 'bla bla!',
        ],
        [
          'term' => 'toto',
          'data_uri' => '',
        ],
        'File type, with term, with config term and no data_uri',
      ],
      [
        'private://genoring/somebase/sorted/bla_bla/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => 'bla bla!',
        ],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'File type, with term, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'directory',
          'toto' => 'bla bla!',
        ],
        [
          'term' => '',
          'data_uri' => '/sorted',
        ],
        'Directory type, with no term, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/bla_bla/',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'directory',
          'toto' => 'bla bla!',
        ],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'Directory type, with term, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
        ],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'File type, with no term, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => '',
        ],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'File type, with empty term, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/sorted/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => ['ab', 'cd'],
        ],
        [
          'term' => 'toto',
          'data_uri' => 'sorted',
        ],
        'File type, with term array, with config term and data_uri',
      ],
      [
        'private://genoring/somebase/multi/terms/cd/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => ['ab', 'cd'],
        ],
        [
          'term' => '$.toto[1]',
          'data_uri' => 'multi/terms/',
        ],
        'File type, with term array, with config json term and data_uri',
      ],
      [
        'private://genoring/somebase/multi/terms/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => ['ab', 'cd'],
        ],
        [
          'term' => '$.toto[2]',
          'data_uri' => 'multi/terms/',
        ],
        'File type, with too short term array, with config json term and data_uri',
      ],
      [
        'private://genoring/somebase/multi/terms/somefile.txt',
        'private://genoring/somebase/somefile.txt',
        [
          'type' => 'file',
          'toto' => 'xyz',
        ],
        [
          'term' => '$.]/-$',
          'data_uri' => 'multi/terms',
        ],
        'File type, with term, with config json term invalid and data_uri',
      ],
    ];
  }

  /**
   * Tests term data locator.
   *
   * @dataProvider provideTestTermDataLocator
   */
  public function testTermDataLocator(
    string $expected_uri,
    string $base_uri,
    array $metadata,
    array $config,
    string $test_name,
  ) {
    $string_translation = $this->createMock(TranslationInterface::class);

    $logger = $this->createMock(LoggerChannelInterface::class);
    $logger_factory = $this->createMock(LoggerChannelFactoryInterface::class);
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('genoring_locator_term')
      ->willReturn($logger);

    $data_locator = new MetadataTermLocator($config, 'term', [], $string_translation, $logger_factory);
    $result_uri = $data_locator->locateUri($metadata, $base_uri);
    $this->assertEquals($expected_uri, $result_uri, $test_name);
  }

}
